# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions


class InternalError(exceptions.NeutronException):
    message = _("Internal mechanism driver error - %(details)s.")

    def __init__(self, **kwargs):
        kwargs.setdefault('details', _("See error log for details"))
        super(InternalError, self).__init__(**kwargs)


class UnsupportedRoutingTopology(exceptions.BadRequest):
    message = _("All router interfaces for a network must share either the "
                "same router or the same subnet.")


class UnscopedSharedNetworkProjectConflict(exceptions.BadRequest):
    message = _("Shared network %(net1)s from project %(proj1)s and shared "
                "network %(net2)s from project %(proj2)s cannot be combined "
                "in the same topology.")


class NonIsomorphicNetworkRoutingUnsupported(exceptions.BadRequest):
    message = _("All router interfaces for a network must utilize the same "
                "VRF.")


class ScopeUpdateNotSupported(exceptions.BadRequest):
    message = _("Updating the address_scope of a subnetpool that is "
                "associated with routers is not currently supported.")


class SnatPortsInUse(exceptions.SubnetInUse):
    def __init__(self, **kwargs):
        kwargs['reason'] = _('Subnet has SNAT IP addresses allocated')
        super(SnatPortsInUse, self).__init__(**kwargs)


class SnatPoolCannotBeUsedForFloatingIp(exceptions.InvalidInput):
    message = _("Floating IP cannot be allocated in SNAT host pool subnet.")


class PreExistingSVICannotBeConnectedToRouter(exceptions.BadRequest):
    message = _("A SVI network with pre-existing l3out is not allowed to "
                "be connected to a router.")


class PreExistingSVICannotUseSameL3out(exceptions.BadRequest):
    message = _("Can not create a SVI network with pre-existing l3out "
                "if that l3out has been used by another SVI network.")


class OnlyOneSubnetInSVINetwork(exceptions.BadRequest):
    message = _("Only one subnet is allowed in SVI network.")


class ExternalSubnetOverlapInL3Out(exceptions.BadRequest):
    message = _("External subnet CIDR %(cidr)s overlaps with existing "
                "subnets in APIC L3Outside %(l3out)s.")


class ExhaustedApicRouterIdPool(exceptions.IpAddressGenerationFailure):
    message = _("All the IPs in the APIC router ID pool %(pool)s "
                "have been taken.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import re
import yaml

# Initialize logging
logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
    level=logging.WARNING)
_log = logging.getLogger()
_log.setLevel(logging.INFO)


orig_getoutput = commands.getoutput


def getoutput(cmd):
    _log.info('Running cmd: %s\n' % (cmd))
    cmd_out = orig_getoutput(cmd)
    _log.info('Cmd output: %s\n' % (cmd_out))
    return cmd_out


commands.getoutput = getoutput


class Gbp_Verify(object):

    def __init__(self):
        """
        Init def
        """
        self.err_strings = [
            'Conflict',
            'Bad Request',
            'Error',
            'Unknown',
            'Unable']

    def gbp_action_verify(self, cmd_val, action_name, *args, **kwargs):
        """
        -- cmd_val== 0:list; 1:show
        -- action_name == UUID or name_string
        List/Show Policy Action
        kwargs addresses the need for passing required/optional params
        """
        if cmd_val == '' or action_name == '':
            _log.info('''Function Usage: gbp_action_verify 0 "abc" \n
                      --cmd_val == 0:list; 1:show\n
                       -- action_name == UUID or name_string\n''')
            return 0
        # Build the command with mandatory param 'action_name'
        if cmd_val == 0:
            cmd = 'gbp policy-action-list | grep %s' % str(action_name)
            for arg in args:
                cmd = cmd + ' | grep %s' % arg
        if cmd_val == 1:
            cmd = "gbp policy-action-show " + str(action_name)

        # Execute the policy-action-verify-cmd
        cmd_out = commands.getoutput(cmd)

        # Catch for non-exception error strings, even though try clause
        # succeded
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(cmd_out)
                _log.info(
                    "Cmd execution failed! with this Return Error: \n%s" %
                    (cmd_out))
                return 0
        if cmd_val == 0:
            for arg in args:
                if cmd_out.find(arg) == -1 or cmd_out.find(action_name) == -1:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s DID NOT MATCH for the Action == %s"
                        "in LIST cmd" % (arg, action_name))
                    return 0
        # If try clause succeeds for "verify" cmd then parse the cmd_out to
        # match the user-fed expected attributes & their values
        if cmd_val == 1:
            for arg, val in kwargs.items():
                if re.search("\\b%s\\b\s+\| \\b%s\\b.*" %
                             (arg, val), cmd_out, re.I) is None:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s and its Value== %s DID NOT MATCH"
                        "for the Action == %s" %
                        (arg, val, action_name))
                    return 0

        return 1

    def gbp_classif_verify(self, cmd_val, classifier_name, *args, **kwargs):
        """
        -- cmd_val== 0:list; 1:show
        -- classifier_name == UUID or name_string
        List/Show Policy Action
        kwargs addresses the need for passing required/optional params
        """
        if cmd_val == '' or classifier_name == '':
            _log.info('''Function Usage: gbp_classif_verify(0,name) \n
                      --cmd_val == 0:list 1:show\n
                       -- classifier_name == UUID or name_string\n''')
            return 0
        # Build the command with mandatory param 'classifier_name'
        if cmd_val == 0:
            cmd = 'gbp policy-classifier-list | grep %s' % str(classifier_name)
            for arg in args:
                cmd = cmd + ' | grep %s' % arg
        if cmd_val == 1:
            cmd = "gbp policy-classifier-show " + str(classifier_name)
        # Execute the policy-classifier-verify-cmd
        cmd_out = commands.getoutput(cmd)

        # Catch for non-exception error strings, even though try clause
        # succeded
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(cmd_out)
                _log.info(
                    "Cmd execution failed! with this Return Error: \n%s" %
                    (cmd_out))
                return 0
        if cmd_val == 0:
            for arg in args:
                if cmd_out.find(arg) == - \
                        1 or cmd_out.find(classifier_name) == -1:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s DID NOT MATCH for the Classifier "
                        "== %s in LIST cmd" %
                        (arg, classifier_name))
                    return 0
        # If try clause succeeds for "verify" cmd then parse the cmd_out to
        # match the user-fed expected attributes & their values
        if cmd_val == 1:
            for arg, val in kwargs.items():
                if re.search("\\b%s\\b\s+\| \\b%s\\b.*" %
                             (arg, val), cmd_out, re.I) is None:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s and its Value== %s DID NOT MATCH "
                        "for the Claasifier == %s" %
                        (arg, val, classifier_name))
                    return 0

        return 1

    def gbp_policy_verify_all(self, cmd_val, verifyobj,
                              name_uuid, *args, **kwargs):
        """
        --verifyobj== policy-*(where *=action;classifer,rule,rule-set,
                      target-group,target)
        --cmd_val== 0:list; 1:show
        kwargs addresses the need for passing required/optional params
        """
        verifyobj_dict = {
            "action": "policy-action",
            "classifier": "policy-classifier",
            "rule": "policy-rule",
            "ruleset": "policy-rule-set",
            "group": "group",
            "target": "policy-target"}
        if verifyobj != '':
            if verifyobj not in verifyobj_dict:
                raise KeyError
        if cmd_val == '' or name_uuid == '':
            _log.info('''Function Usage: gbp_policy_verify_all(0,'action',
                      'name_uuid')\n
                      --cmd_val == 0:list; 1:show\n
                      -- name_uuid == UUID or name_string\n''')
            return 0
        # Build the command with mandatory params
        if cmd_val == 0:
            cmd = 'gbp %s-list | grep ' % verifyobj_dict[
                verifyobj] + str(name_uuid)
            for arg in args:
                cmd = cmd + ' | grep %s' % arg
        if cmd_val == 1:
            cmd = 'gbp %s-show ' % verifyobj_dict[verifyobj] + str(name_uuid)
        # Execute the policy-object-verify-cmd
        cmd_out = commands.getoutput(cmd)
        # Catch for non-exception error strings
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(cmd_out)
                _log.info(
                    "Cmd execution failed! with this Return Error: \n%s" %
                    (cmd_out))
                return 0
        if cmd_val == 0:
            if cmd_out == '':  # The case when grep returns null
                return 0
            else:
                for arg in args:
                    if (cmd_out.find(arg) == -1 or
                            cmd_out.find(name_uuid) == -1):
                        _log.info(cmd_out)
                        _log.info(
                            "The Attribute== %s DID NOT MATCH for the "
                            "Policy Object == %s in LIST cmd" %
                            (arg, verifyobj))
                        return 0
        # If "verify" cmd succeeds then parse the cmd_out to match the user-fed
        # expected attributes & their values
        if cmd_val == 1:
            for arg, val in kwargs.items():
                if re.search("\\b%s\\b\s+\| \\b%s\\b.*" %
                             (arg, val), cmd_out, re.I) is None:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s and its Value== %s DID NOT MATCH "
                        "for the PolicyObject == %s" %
                        (arg, val, verifyobj))
                    return 0

        return 1

    def gbp_l2l3ntk_pol_ver_all(
            self, cmd_val, verifyobj, name_uuid, ret='', *args, **kwargs):
        """
        --verifyobj== *policy(where *=l2;l3,network)
        --cmd_val== 0:list; 1:show
        --ret=='default' <<< function will return some attribute values
        depending upon the verifyobj
        kwargs addresses the need for passing required/optional params
        """
        verifyobj_dict = {
            "l2p": "l2policy",
            "l3p": "l3policy",
            "nsp": "network-service-policy"}
        if verifyobj != '':
            if verifyobj not in verifyobj_dict:
                raise KeyError
        if cmd_val == '' or name_uuid == '':
            _log.info('''Function Usage: gbp_l2l3ntk_pol_ver_all(0,
                      'l2p','name') \n
                      --cmd_val == 0:list; 1:show\n
                      --name_uuid == UUID or name_string\n''')
            return 0
        # Build the command with mandatory params
        if cmd_val == 0:
            cmd = 'gbp %s-list | grep ' % verifyobj_dict[
                verifyobj] + str(name_uuid)
            for arg in args:
                cmd += ' | grep %s' % arg
        if cmd_val == 1:
            cmd = 'gbp %s-show ' % verifyobj_dict[verifyobj] + str(name_uuid)
        # Execute the policy-object-verify-cmd
        cmd_out = commands.getoutput(cmd)
        # _log.info(cmd_out)
        # Catch for non-exception error strings
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(cmd_out)
                _log.info(
                    "Cmd execution failed! with this Return Error: \n%s" %
                    (cmd_out))
                return 0
        if cmd_val == 0:
            if cmd_out == '':  # The case when grep returns null
                return 0
            else:
                for arg in args:
                    if (cmd_out.find(arg) == -1 or
                            cmd_out.find(name_uuid) == -1):
                        _log.info(cmd_out)
                        _log.info(
                            "The Attribute== %s DID NOT MATCH for the Policy "
                            "Object == %s in LIST cmd" %
                            (arg, verifyobj))
                        return 0
        # If "verify" succeeds cmd then parse the cmd_out to match the user-fed
        # expected attributes & their values
        if cmd_val == 1 and ret == 'default':
            for arg, val in kwargs.items():
                if re.search("\\b%s\\b\s+\| \\b%s\\b.*" %
                             (arg, val), cmd_out, re.I) is None:
                    # incase of attribute has more than one value then
                    # then below function will help us validating the values
                    # or the only value among all for the given attr.
                    # Example: L2P can have multiple PTGs, L3P can have multi
                    # L2Ps
                    if not self.gbp_obj_ver_attr_all_values(
                            verifyobj, name_uuid, arg, [val]):
                        _log.info(
                            "The Attribute== %s and its Value== %s "
                            "DID NOT MATCH "
                            "for the PolicyObject == %s" % (
                                arg, val, verifyobj))
                        return 0
            if verifyobj == "l2p":
                match = re.search(
                    "\\bl3_policy_id\\b\s+\| (.*) \|", cmd_out, re.I)
                l3pid = match.group(1)
                match = re.search(
                    "\\bnetwork_id\\b\s+\| (.*) \|", cmd_out, re.I)
                ntkid = match.group(1)
                return l3pid.rstrip(), ntkid.rstrip()
            if verifyobj == "l3p":
                match = re.search("\\brouters\\b\s+\| (.*) \|", cmd_out, re.I)
                rtrid = match.group(1)
                return rtrid.rstrip()
        elif cmd_val == 1:
            for arg, val in kwargs.items():
                if arg == 'network_service_params':
                    if re.findall('(%s)' % (val), cmd_out) == []:
                        _log.info(cmd_out)
                        _log.info(
                            "The Attribute== %s and its Value== %s DID NOT "
                            "MATCH for the PolicyObject == %s" %
                            (arg, val, verifyobj))
                        return 0
                elif re.search("\\b%s\\b\s+\| \\b%s\\b.*" % (arg, val),
                               cmd_out, re.I) is None:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s and its Value== %s DID NOT MATCH "
                        "for the PolicyObject == %s" %
                        (arg, val, verifyobj))
                    return 0
        else:
            return 1

    def neut_ver_all(self, verifyobj, name_uuid, ret='', **kwargs):
        """
        --verifyobj== net,subnet,port,router
        --ret=='default' <<< function will return some attribute values
                depending upon the verifyobj
        kwargs addresses the need for passing required/optional params
        """
        if name_uuid == '':
            _log.info('''Function Usage: neut_ver_all('net','name')\n
                      -- name_uuid == UUID or name_string\n''')
            return 0
        # Build the command with mandatory params
        cmd = 'neutron %s-show ' % verifyobj + str(name_uuid)
        _log.info('Neutron Cmd == %s\n' % (cmd))
        # Execute the policy-object-verify-cmd
        cmd_out = commands.getoutput(cmd)
        _log.info(cmd_out)
        # Catch for non-exception error strings
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(cmd_out)
                _log.info(
                    "Neutron Cmd execution failed! with this Return Error: "
                    "\n%s" % cmd_out)
                return 0
        if ret != '':
            match = re.search("\\b%s\\b\s+\| (.*) \|" % (ret), cmd_out, re.I)
            if match is not None:
                return match.group(1).rstrip()
            else:
                return 0
        for arg, val in kwargs.items():
            if isinstance(val, list):  # More than 1 value is to be verified
                for i in val:
                    if cmd_out.find(i) == -1:
                        _log.info(cmd_out)
                        _log.info(
                            "The Attribute== %s and its Value== %s DID NOT "
                            "MATCH for the NeutronObject == %s" %
                            (arg, i, verifyobj))
                        return 0
            else:
                if re.search("\\b%s\\b\s+\| \\b%s\\b.*" %
                             (arg, val), cmd_out, re.I) is None:
                    _log.info(cmd_out)
                    _log.info(
                        "The Attribute== %s and its Value== %s DID NOT MATCH "
                        "for the NeutronObject == %s" %
                        (arg, val, verifyobj))
                    return 0
        return 1

    def gbp_obj_ver_attr_all_values(self, verifyobj, name_uuid, attr, values):
        """
        Function will verify multiple entries for any given attribute
        of a Policy Object
        --values=Must be a list
        """
        verifyobj_dict = {
            "action": "policy-action",
            "classifier": "policy-classifier",
            "rule": "policy-rule",
            "ruleset": "policy-rule-set",
            "group": "group",
            "target": "policy-target",
            "l2p": "l2policy",
            "l3p": "l3policy",
            "nsp": "network-service-policy"}
        if verifyobj != '':
            if verifyobj not in verifyobj_dict:
                raise KeyError
        if not isinstance(values, list):
            raise TypeError
        # Build the command with mandatory params
        cmd = ('gbp %s-show ' % verifyobj_dict[verifyobj] +
               str(name_uuid) + ' -F %s' % (attr))
        # Execute the policy-object-verify-cmd
        cmd_out = commands.getoutput(cmd)
        # Catch for non-exception error strings
        for err in self.err_strings:
            if re.search('\\b%s\\b' % (err), cmd_out, re.I):
                _log.info(
                    "Cmd execution failed! with this Return Error: \n%s" %
                    (cmd_out))
                return 0
        _misses = []
        for val in values:
            if cmd_out.find(val) == -1:
                _misses.append(val)
        if len(_misses) > 0:
            _log.info(
                "\nFollowing Values of the Attribute for the Policy Object "
                "was NOT FOUND=%s" %
                (_misses))
            return 0
        return 1

    def get_uuid_from_stack(self, yaml_file, heat_stack_name):
        """
        Fetches the UUID of the GBP Objects created by Heat
        """
        with open(yaml_file, 'rt') as f:
            heat_conf = yaml.load(f)
        obj_uuid = {}
        # This comprise dict with keys as in [outputs] block of yaml-based
        # heat template
        outputs_dict = heat_conf["outputs"]
        print(outputs_dict)
        for key in outputs_dict.iterkeys():
            cmd = 'heat stack-show %s | grep -B 2 %s' % (heat_stack_name, key)
            print(cmd)
            cmd_out = commands.getoutput(cmd)
            print(cmd_out)
            match = re.search('\"\\boutput_value\\b\": \"(.*)\"',
                              cmd_out, re.I)
            if match is not None:
                obj_uuid[key] = match.group(1)
        return obj_uuid



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aim.api import infra as aim_infra
from aim.api import resource as aim_res
from aim.api import service_graph as aim_sg
import mock
from networking_sfc.extensions import flowclassifier as flowc_ext
from networking_sfc.extensions import sfc as sfc_ext
from networking_sfc.services.flowclassifier.common import config as flc_cfg
from networking_sfc.services.flowclassifier import driver_manager as fc_driverm
from networking_sfc.services.sfc.common import config as sfc_cfg
from networking_sfc.services.sfc import driver_manager as sfc_driverm
from neutron.callbacks import exceptions as c_exc
from neutron import context
from neutron.db import l3_db
from neutron import manager
from oslo_log import log as logging

from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.tests.unit.db.grouppolicy import test_group_policy_db
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_aim_mapping_driver as test_aim_base)

LOG = logging.getLogger(__name__)


class Rollback(Exception):
    pass


class TestAIMServiceFunctionChainingBase(test_aim_base.AIMBaseTestCase):

    def setUp(self, *args, **kwargs):
        sfc_cfg.cfg.CONF.set_override('drivers', ['aim'], group='sfc')
        flc_cfg.cfg.CONF.set_override('drivers', ['aim'],
                                      group='flowclassifier')
        config.cfg.CONF.set_override(
            'network_vlan_ranges', ['physnet1:100:200'], group='ml2_type_vlan')
        ml2_options = {'mechanism_drivers': ['apic_aim', 'openvswitch'],
                       'extension_drivers': ['apic_aim', 'port_security',
                                             'dns'],
                       'type_drivers': ['opflex', 'local', 'vlan'],
                       'tenant_network_types': ['vlan']}
        # NOTE(ivar): the SFC and FLC driver managers load the driver names in
        # the default parameters of their INIT functions. In Python, default
        # params are evaluated only once when the module is loaded hence
        # causing issues in the tests if those modules ever get loaded before
        # the aim override happens. We need to reload the modules at this point
        # to fix the issue.
        reload(fc_driverm)
        reload(sfc_driverm)
        super(TestAIMServiceFunctionChainingBase, self).setUp(
            *args, ml2_options=ml2_options, trunk_plugin='trunk', **kwargs)
        self.agent_conf = test_group_policy_db.AGENT_CONF
        self._sfc_driver = None
        self._flowc_driver = None
        self._sfc_plugin = None
        self._flowc_plugin = None
        self._aim_mech_driver = None
        self.hlink1 = aim_infra.HostLink(host_name='h1', interface_name='eth0',
                           path='topology/pod-1/paths-101/pathep-[eth1/1]')
        self.hlink2 = aim_infra.HostLink(host_name='h2', interface_name='eth0',
                           path='topology/pod-1/paths-102/pathep-[eth1/1]')
        self.path_by_host = {'h1': 'topology/pod-1/paths-101/pathep-[eth1/1]',
                             'h2': 'topology/pod-1/paths-102/pathep-[eth1/1]'}
        self.physdom = aim_res.PhysicalDomain(name='sfc-phys', monitored=True)

        self.aim_mgr.create(self._aim_context, self.hlink1)
        self.aim_mgr.create(self._aim_context, self.hlink2)
        self.aim_mgr.create(self._aim_context, self.physdom)
        self.aim_mgr.create(self._aim_context,
                            aim_infra.HostDomainMappingV2(
                                host_name='h1', domain_name=self.physdom.name,
                                domain_type='PhysDom'))
        self.aim_mgr.create(self._aim_context,
                            aim_infra.HostDomainMappingV2(
                                host_name='h2', domain_name=self.physdom.name,
                                domain_type='PhysDom'))
        self._plugin = manager.NeutronManager.get_plugin()
        self._plugin.remove_networks_from_down_agents = mock.Mock()
        self._plugin.is_agent_down = mock.Mock(return_value=False)
        self._ctx = context.get_admin_context()

    def tearDown(self):
        LOG.warning("SFCDs used in this test: %s",
                    self.sfc_plugin.driver_manager.drivers.keys())
        LOG.warning("FLCDs used in this test: %s",
                    self.flowc_plugin.driver_manager.drivers.keys())
        # Always reset configuration to dummy driver. Any
        # test which requires to configure a different
        # policy driver would have done so in it's setup
        # (and should have ideally reset it too).
        config.cfg.CONF.set_override('drivers', ['dummy'], group='sfc')
        config.cfg.CONF.set_override('drivers', ['dummy'],
                                     group='flowclassifier')
        super(TestAIMServiceFunctionChainingBase, self).tearDown()

    @property
    def sfc_plugin(self):
        if not self._sfc_plugin:
            plugins = manager.NeutronManager.get_service_plugins()
            self._sfc_plugin = plugins.get(sfc_ext.SFC_EXT)
        return self._sfc_plugin

    @property
    def flowc_plugin(self):
        if not self._flowc_plugin:
            plugins = manager.NeutronManager.get_service_plugins()
            self._flowc_plugin = plugins.get(flowc_ext.FLOW_CLASSIFIER_EXT)
        return self._flowc_plugin

    @property
    def sfc_driver(self):
        # aim_mapping policy driver reference
        if not self._sfc_driver:
            self._sfc_driver = (
                self.sfc_plugin.driver_manager.drivers['aim'].obj)
        return self._sfc_driver

    @property
    def flowc_driver(self):
        # aim_mapping policy driver reference
        if not self._flowc_driver:
            self._flowc_driver = (
                self.flowc_plugin.driver_manager.drivers['aim'].obj)
        return self._flowc_driver

    @property
    def aim_mech(self):
        if not self._aim_mech_driver:
            self._aim_mech_driver = (
                self._plugin.mechanism_manager.mech_drivers['apic_aim'].obj)
        return self._aim_mech_driver

    def _create_simple_ppg(self, pairs=2, leftn_id=None, rightn_id=None,
                           check_type=None, check_freq=None, check_port=None):
        nets = []
        # Pairs go in 2 networks
        if not leftn_id or not rightn_id:
            for i in range(2):
                net = self._make_network(self.fmt, 'net1', True)
                self._make_subnet(self.fmt, net, '192.168.%s.1' % i,
                                  '192.168.%s.0/24' % i)
                nets.append(net['network']['id'])
        else:
            nets = [leftn_id, rightn_id]

        port_pairs = []
        for i in range(pairs):
            p1 = self._make_port(self.fmt, nets[0])['port']
            self._bind_port_to_host(p1['id'], 'h%s' % ((i % 2) + 1))
            self._plugin.update_port_status(self._ctx, p1['id'], 'ACTIVE')
            p2 = self._make_port(self.fmt, nets[1])['port']
            self._bind_port_to_host(p2['id'], 'h%s' % ((i % 2) + 1))
            self._plugin.update_port_status(self._ctx, p2['id'], 'ACTIVE')
            pp = self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                                       expected_res_status=201)['port_pair']
            port_pairs.append(pp)
        # This goes through
        kwargs = {}
        port_pair_group_parameters = {}
        if check_type:
            port_pair_group_parameters['healthcheck_type'] = check_type
        if check_freq:
            port_pair_group_parameters['healthcheck_frequency'] = check_freq
        if check_port:
            port_pair_group_parameters['healthcheck_tcp_port'] = check_port
        if port_pair_group_parameters:
            kwargs['port_pair_group_parameters'] = port_pair_group_parameters
        return self.create_port_pair_group(
            port_pairs=[pp['id'] for pp in port_pairs],
            expected_res_status=201, **kwargs)['port_pair_group']

    def _create_simple_flowc(self, src_svi=False, dst_svi=False):
        kwargs = {}

        def get_svi_kwargs():
            return {'apic:svi': True}

        # Need to make sure all the SVI networks are in the same common
        # unrouted vrf as other service chain resources do.
        if src_svi or dst_svi:
            router = self._make_router(
                self.fmt, self._tenant_id, 'router1')['router']
            vrf_name = 'openstack_UnroutedVRF'
            vrf = aim_res.VRF(tenant_name='common', name=vrf_name,
                              monitored=True)
            as1 = self._make_address_scope_for_vrf(
                vrf.dn, name='as1')['address_scope']
            pool = self._make_subnetpool(
                self.fmt, ['192.168.0.0/8'], name='sp',
                address_scope_id=as1['id'], tenant_id=as1['tenant_id'],
                default_prefixlen=24)['subnetpool']
        if src_svi:
            # We need to create the L3Out and the External network
            kwargs = get_svi_kwargs()
        net1 = self._make_network(self.fmt, 'net1', True,
                                  arg_list=self.extension_attributes,
                                  **kwargs)
        if src_svi:
            subnet1 = self._make_subnet(
                self.fmt, net1, '192.168.0.1', '192.168.0.0/24',
                subnetpool_id=pool['id'])['subnet']
            self.l3_plugin.add_router_interface(
                context.get_admin_context(), router['id'],
                {'subnet_id': subnet1['id']})
        else:
            self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        kwargs = {}
        if dst_svi:
            kwargs = get_svi_kwargs()
        net2 = self._make_network(self.fmt, 'net2', True,
                                  arg_list=self.extension_attributes,
                                  **kwargs)
        if dst_svi:
            subnet2 = self._make_subnet(
                self.fmt, net2, '192.168.1.1', '192.168.1.0/24',
                subnetpool_id=pool['id'])['subnet']
            self.l3_plugin.add_router_interface(
                context.get_admin_context(), router['id'],
                {'subnet_id': subnet2['id']})
        else:
            self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')

        return self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=201)['flow_classifier']

    def _create_simple_port_chain(self, flowcs=1, ppgs=2, flowcs_args=None,
                                  ppg_args=None):
        flowc_ids = []
        ppg_args = ppg_args or []
        flowcs_args = flowcs_args or []
        for i in range(flowcs):
            try:
                flowc_ids.append(
                    self._create_simple_flowc(**flowcs_args[i])['id'])
            except IndexError:
                flowc_ids.append(self._create_simple_flowc()['id'])
        ppg_ids = []
        for i in range(ppgs):
            try:
                ppg_ids.append(self._create_simple_ppg(**ppg_args[i])['id'])
            except IndexError:
                ppg_ids.append(self._create_simple_ppg()['id'])
        return self.create_port_chain(port_pair_groups=ppg_ids,
                                      flow_classifiers=flowc_ids,
                                      expected_res_status=201)['port_chain']

    def _verify_ppg_mapping(self, ppg, tenant):
        apic_tn = tenant
        # Verify expected AIM model
        ctx = self._aim_context
        # DeviceCluster. Only one created
        dc = self.aim_mgr.get(ctx, aim_sg.DeviceCluster(
            tenant_name=apic_tn, name='ppg_' + ppg['id']))
        ppg_params = ppg['port_pair_group_parameters']
        hcheck_type = ppg_params.get('healthcheck_type')
        if hcheck_type:
            hcheck_frequency = ppg_params.get('healthcheck_frequency')
            hcheck_tcp_port = ppg_params.get('healthcheck_tcp_port')
        mp = self.aim_mgr.get(ctx, aim_sg.ServiceRedirectMonitoringPolicy(
            tenant_name=apic_tn, name='ppg_' + ppg['id']))
        if hcheck_type:
            self.assertIsNotNone(mp)
            self.assertEqual(hcheck_type, mp.type)
            if hcheck_frequency:
                self.assertEqual(str(hcheck_frequency), mp.frequency)
            if hcheck_tcp_port:
                self.assertEqual(str(hcheck_tcp_port), mp.tcp_port)
        else:
            # Test hcheck toggle
            self.assertIsNone(mp)

        self.assertIsNotNone(dc)
        self.assertEqual(self.physdom.name, dc.physical_domain_name)
        pps = [self.show_port_pair(x)['port_pair'] for x in ppg['port_pairs']]

        srgh_dn_by_pp = {}
        for pp in pps:
            self.assertIsNotNone(self.aim_mgr.get(
                ctx, aim_sg.ConcreteDevice(tenant_name=apic_tn,
                                           device_cluster_name=dc.name,
                                           name='pp_' + pp['id'])))
            srhg = aim_sg.ServiceRedirectHealthGroup(
                tenant_name=dc.tenant_name, name='pp_' + pp['id'])
            srgh_dn_by_pp[pp['id']] = srhg.dn
            srhg = self.aim_mgr.get(ctx, srhg)
            if hcheck_type:
                self.assertIsNotNone(srhg)
            else:
                self.assertIsNone(srhg)

        for pp in pps:
            # Each of these CD have 2 CDIs
            iprt = self._show_port(pp['ingress'])
            eprt = self._show_port(pp['egress'])
            for p in [iprt, eprt]:
                pp_dciin = self.aim_mgr.find(
                    ctx, aim_sg.ConcreteDeviceInterface,
                    name='prt_' + p['id'])
                pp_dciin = pp_dciin[0]
                self.assertEqual(
                    self.path_by_host.get(p['binding:host_id'], ''),
                    pp_dciin.path)
            iepg = self.aim_mech._get_epg_by_network_id(self._ctx.session,
                                                        iprt['network_id'])
            eepg = self.aim_mech._get_epg_by_network_id(self._ctx.session,
                                                        eprt['network_id'])
            self.assertFalse(self.aim_mgr.get(ctx, iepg).sync)
            self.assertFalse(self.aim_mgr.get(ctx, eepg).sync)

        # 1 PPG means 1 service, which has 2 DeviceClusterInterfaces
        # comprehensive of all the above ConcreteDeviceInterfaces
        idci = self.aim_mgr.get(ctx, aim_sg.DeviceClusterInterface(
            tenant_name=dc.tenant_name, device_cluster_name=dc.name,
            name='ingress'))
        edci = self.aim_mgr.get(ctx, aim_sg.DeviceClusterInterface(
            tenant_name=dc.tenant_name, device_cluster_name=dc.name,
            name='egress'))
        self.assertIsNotNone(idci)
        self.assertIsNotNone(edci)

        # Ingress CDIs
        ingr_cdis = [
            aim_sg.ConcreteDeviceInterface(
                tenant_name=apic_tn, device_cluster_name=dc.name,
                device_name='pp_' + pp['id'], name='prt_' + pp['ingress'])
            for pp in pps]

        self.assertEqual({ingr.dn for ingr in ingr_cdis},
                         set(idci.concrete_interfaces))

        # Egress CDIs
        egr_cdis = [
            aim_sg.ConcreteDeviceInterface(
                tenant_name=apic_tn, device_cluster_name=dc.name,
                device_name='pp_' + pp['id'], name='prt_' + pp['egress'])
            for pp in pps]

        self.assertEqual({egr.dn for egr in egr_cdis},
                         set(edci.concrete_interfaces))
        # Redirect Policy Ingress
        irp = self.aim_mgr.get(ctx, aim_sg.ServiceRedirectPolicy(
            tenant_name=dc.tenant_name, name='ingr_ppg_' + ppg['id']))
        erp = self.aim_mgr.get(ctx, aim_sg.ServiceRedirectPolicy(
            tenant_name=dc.tenant_name, name='egr_ppg_' + ppg['id']))
        self.assertIsNotNone(irp)
        self.assertIsNotNone(erp)

        # Ingress Ports
        for type, pbr in [('ingress', irp), ('egress', erp)]:
            prts = [(self._show_port(pp[type]), pp) for pp in pps]
            observed_destinations = []
            for port, pp in prts:
                dst = {'ip': port['fixed_ips'][0]['ip_address'],
                       'mac': port['mac_address'], 'name': 'pp_' + pp['id']}
                if hcheck_type:
                    dst['redirect_health_group_dn'] = srgh_dn_by_pp[pp['id']]
                observed_destinations.append(dst)
            self.assertEqual(sorted(observed_destinations), pbr.destinations)

    def _verify_pc_mapping(self, pc, multiple=False):
        ctx = self._aim_context
        flowcs = [self.show_flow_classifier(x)['flow_classifier'] for x in
                  pc['flow_classifiers']]
        flowc_tenants = set([self._show_network(
            flowc['l7_parameters']['logical_destination_network'])['tenant_id']
            for flowc in flowcs])
        ppgs = [self.show_port_pair_group(x)['port_pair_group'] for x in
                pc['port_pair_groups']]
        if not multiple:
            self.assertEqual(
                len(flowc_tenants) * len(ppgs),
                len(self.aim_mgr.find(ctx, aim_sg.DeviceClusterContext)))
            self.assertEqual(
                len(flowc_tenants) * len(ppgs) * 2,
                len(self.aim_mgr.find(ctx,
                                      aim_sg.DeviceClusterInterfaceContext)))
            self.assertEqual(
                len(flowc_tenants),
                len(self.aim_mgr.find(ctx, aim_sg.ServiceGraph)))
        for flowc in flowcs:
            src_net = self._show_network(
                flowc['l7_parameters']['logical_source_network'])
            dst_net = self._show_network(
                flowc['l7_parameters']['logical_destination_network'])
            apic_tn = 'prj_' + dst_net['tenant_id']
            device_clusters = []
            sg = self.aim_mgr.get(ctx, aim_sg.ServiceGraph(
                tenant_name=apic_tn, name='ptc_' + pc['id']))
            self.assertIsNotNone(sg)
            src_cidr = flowc['source_ip_prefix']
            dst_cird = flowc['destination_ip_prefix']

            def get_net_group(net, cidr):
                if net['apic:svi']:
                    # TODO(ivar): this will not work, there's no L3Outside
                    # DN extension for external networks.
                    ext = aim_res.ExternalNetwork.from_dn(
                        net['apic:distinguished_names']['ExternalNetwork'])
                    name_prefix = cidr.replace('/', '_')
                    if cidr in ['0.0.0.0/0', '::/0']:
                        # use default external EPG
                        name_prefix = 'default'
                    ext_net = self.aim_mgr.get(
                        ctx, aim_res.ExternalNetwork(
                            tenant_name=ext.tenant_name,
                            l3out_name=ext.l3out_name,
                            name=name_prefix + '_' + 'net_' + net['id']))
                    return ext_net
                else:
                    epg = self.aim_mgr.get(
                        ctx, aim_res.EndpointGroup.from_dn(
                            net['apic:distinguished_names']['EndpointGroup']))
                    return epg

            provider = get_net_group(dst_net, dst_cird)
            # Verify Flow Classifier mapping
            contract = self.aim_mgr.get(
                ctx, aim_res.Contract(
                    tenant_name=apic_tn,
                    name=self.sfc_driver._generate_contract_name(provider.name,
                                                                 sg.name)))
            self.assertIsNotNone(contract)
            subject = self.aim_mgr.get(
                ctx, aim_res.ContractSubject(
                    tenant_name=apic_tn,
                    contract_name=contract.name,
                    name='ptc_' + pc['id']))
            self.assertIsNotNone(subject)
            self.assertEqual(['openstack_AnyFilter'], subject.bi_filters)
            for net, pref, cidr in [(src_net, 'src_', src_cidr),
                                    (dst_net, 'dst_', dst_cird)]:
                group = get_net_group(net, cidr)
                if net['apic:svi']:
                    ext_net = group
                    subnets = [cidr]
                    if cidr in ['0.0.0.0/0', '::/0']:
                        # use default external EPG
                        subnets = ['128.0.0.0/1', '0.0.0.0/1', '8000::/1',
                                   '::/1']
                    for sub in subnets:
                        ext_sub = self.aim_mgr.get(ctx, aim_res.ExternalSubnet(
                            tenant_name=ext_net.tenant_name,
                            l3out_name=ext_net.l3out_name,
                            external_network_name=ext_net.name, cidr=sub))
                        self.assertIsNotNone(ext_sub)

                    self.assertIsNotNone(ext_net)
                    self.assertTrue(
                        contract.name in (ext_net.consumed_contract_names if
                                          pref == 'src_' else
                                          ext_net.provided_contract_names),
                        "%s not in ext net %s" % (contract.name,
                                                  ext_net.__dict__))
                else:
                    epg = group
                    self.assertTrue(
                        contract.name in (epg.consumed_contract_names if
                                          pref == 'src_' else
                                          epg.provided_contract_names))
            for ppg in ppgs:
                self._verify_ppg_mapping(ppg, apic_tn)
                device_cluster = self.aim_mgr.get(
                    ctx, aim_sg.DeviceCluster(tenant_name=apic_tn,
                                              name='ppg_' + ppg['id']))
                device_clusters.append(device_cluster)
                dcc = self.aim_mgr.get(
                    ctx, aim_sg.DeviceClusterContext(
                        tenant_name=sg.tenant_name,
                        contract_name="any",
                        service_graph_name=sg.name,
                        node_name=device_cluster.name))
                self.assertIsNotNone(dcc)
                self.assertEqual(device_cluster.name, dcc.device_cluster_name)
                self.assertEqual(apic_tn, dcc.device_cluster_tenant_name)
                # Get ingress/egress BD
                pp = self.show_port_pair(ppg['port_pairs'][0])['port_pair']
                ingress_net = self._get_port_network(pp['ingress'])
                egress_net = self._get_port_network(pp['egress'])
                ingress_bd = ingress_net[
                    'apic:distinguished_names']['BridgeDomain']
                egress_bd = egress_net[
                    'apic:distinguished_names']['BridgeDomain']

                dci = aim_sg.DeviceClusterInterface(
                    tenant_name=device_cluster.tenant_name,
                    device_cluster_name=device_cluster.name, name='ingress')
                dcic = aim_sg.DeviceClusterInterfaceContext(
                    tenant_name=apic_tn, contract_name="any",
                    service_graph_name=sg.name, node_name=device_cluster.name,
                    connector_name='consumer')
                dcic = self.aim_mgr.get(ctx, dcic)
                self.assertIsNotNone(dcic)
                self.assertEqual(ingress_bd, dcic.bridge_domain_dn)
                self.assertEqual(dci.dn, dcic.device_cluster_interface_dn)
                self.assertNotEqual('', dcic.service_redirect_policy_dn)

                dci = aim_sg.DeviceClusterInterface(
                    tenant_name=device_cluster.tenant_name,
                    device_cluster_name=device_cluster.name, name='egress')
                dcic = aim_sg.DeviceClusterInterfaceContext(
                    tenant_name=apic_tn, contract_name="any",
                    service_graph_name=sg.name, node_name=device_cluster.name,
                    connector_name='provider')
                dcic = self.aim_mgr.get(ctx, dcic)
                self.assertIsNotNone(dcic)
                self.assertEqual(egress_bd, dcic.bridge_domain_dn)
                self.assertEqual(dci.dn, dcic.device_cluster_interface_dn)
                self.assertNotEqual('', dcic.service_redirect_policy_dn)
            self.assertEqual(
                sorted({'name': x.name, 'device_cluster_name': x.name,
                        'device_cluster_tenant_name': x.tenant_name}
                       for x in device_clusters),
                sorted(sg.linear_chain_nodes))

    def _verify_pc_delete(self, pc):
        ctx = self._aim_context
        self.delete_port_chain(pc['id'])
        # PC and Flowc unmapped
        routers_count = ctx.db_session.query(l3_db.Router).count()
        self.assertEqual(routers_count,
                         len(self.aim_mgr.find(ctx, aim_res.Contract)))
        self.assertEqual(routers_count,
                         len(self.aim_mgr.find(ctx, aim_res.ContractSubject)))
        self.assertEqual(
            [], self.aim_mgr.find(ctx, aim_sg.DeviceClusterContext))
        self.assertEqual(
            [], self.aim_mgr.find(ctx, aim_sg.DeviceClusterInterfaceContext))
        self.assertEqual([], self.aim_mgr.find(ctx, aim_sg.ServiceGraph))
        # PPGs unmapped
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.ServiceRedirectPolicy)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.ConcreteDeviceInterface)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.ConcreteDevice)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.DeviceCluster)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.DeviceClusterInterface)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx,
                                     aim_sg.ServiceRedirectMonitoringPolicy)))
        self.assertEqual(
            0, len(self.aim_mgr.find(ctx, aim_sg.ServiceRedirectHealthGroup)))

        ppgs = [self.show_port_pair_group(x)['port_pair_group'] for x in
                pc['port_pair_groups']]
        for ppg in ppgs:
            pps = [self.show_port_pair(x)['port_pair'] for x in
                   ppg['port_pairs']]
            for pp in pps:
                iprt = self._show_port(pp['ingress'])
                eprt = self._show_port(pp['egress'])
                iepg = self.aim_mech._get_epg_by_network_id(self._ctx.session,
                                                            iprt['network_id'])
                eepg = self.aim_mech._get_epg_by_network_id(self._ctx.session,
                                                            eprt['network_id'])
                self.assertTrue(self.aim_mgr.get(ctx, iepg).sync)
                self.assertTrue(self.aim_mgr.get(ctx, eepg).sync)

    def _delete_network(self, network_id):
        req = self.new_delete_request('networks', network_id)
        return req.get_response(self.api)


class TestPortPair(TestAIMServiceFunctionChainingBase):

    def test_port_pair_validation(self):
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        p1 = self._make_port(self.fmt, net1['network']['id'])['port']

        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net1, '192.168.1.1', '192.168.1.0/24')
        p2 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p1['id'], 'h1')
        self._bind_port_to_host(p2['id'], 'h2')
        self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                              expected_res_status=201)
        # Same network ports
        p3 = self._make_port(self.fmt, net2['network']['id'])['port']
        p4 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p3['id'], 'h1')
        self._bind_port_to_host(p4['id'], 'h2')
        self.create_port_pair(ingress=p3['id'], egress=p4['id'],
                              expected_res_status=500)
        # Also unbound ports can be used
        p5 = self._make_port(self.fmt, net1['network']['id'])['port']
        self.create_port_pair(ingress=p3['id'], egress=p5['id'],
                              expected_res_status=400)
        # Ports with no domain

    def test_port_pair_validation_no_domain(self):
        self.aim_mgr.delete(self._aim_context, self.physdom)
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net1, '192.168.1.1', '192.168.1.0/24')
        p1 = self._make_port(self.fmt, net1['network']['id'])['port']
        p2 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p1['id'], 'h1')
        # H3 has no domain specified
        self._bind_port_to_host(p2['id'], 'h3')
        self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                              expected_res_status=500)
        # Both ports no domain
        p3 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p3['id'], 'h4')
        self.create_port_pair(ingress=p3['id'], egress=p2['id'],
                              expected_res_status=500)
        # Add domain, but different than H1
        pd = self.aim_mgr.create(
            self._aim_context, aim_infra.HostDomainMappingV2(
                host_name='h3', domain_name='diff-name',
                domain_type='PhysDom'))
        self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                              expected_res_status=500)
        # Multi domain per host
        self.aim_mgr.create(self._aim_context, aim_infra.HostDomainMappingV2(
            host_name='h3', domain_name=self.physdom.name,
            domain_type='PhysDom'))
        self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                              expected_res_status=500)
        # Delete extra domain
        self.aim_mgr.delete(self._aim_context, pd)
        self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                              expected_res_status=201)

    def test_port_pair_validation_trunk(self):
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        snet1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.167.0.1', '192.167.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net1, '192.168.1.1', '192.168.1.0/24')
        snet2 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.167.1.1', '192.167.1.0/24')
        p1 = self._make_port(self.fmt, net1['network']['id'])['port']
        p2 = self._make_port(self.fmt, net2['network']['id'])['port']
        sp1 = self._make_port(self.fmt, snet1['network']['id'])['port']
        sp2 = self._make_port(self.fmt, snet2['network']['id'])['port']
        trunk1 = self._create_resource('trunk', port_id=p1['id'])
        trunk2 = self._create_resource('trunk', port_id=p2['id'])
        self._bind_port_to_host(p1['id'], 'h1')
        self._bind_port_to_host(p2['id'], 'h2')
        self._bind_subport(self._ctx, trunk1, sp1)
        self._bind_subport(self._ctx, trunk2, sp2)
        self.driver._trunk_plugin.add_subports(
            self._ctx, trunk1['trunk']['id'],
            {'sub_ports': [{'port_id': sp1['id'],
                            'segmentation_type': 'vlan',
                            'segmentation_id': 100}]})
        self.driver._trunk_plugin.add_subports(
            self._ctx, trunk2['trunk']['id'],
            {'sub_ports': [{'port_id': sp2['id'],
                            'segmentation_type': 'vlan',
                            'segmentation_id': 100}]})
        self.create_port_pair(ingress=sp1['id'], egress=sp2['id'],
                              expected_res_status=201)


class TestPortPairGroup(TestAIMServiceFunctionChainingBase):

    def test_ppg_validation(self):
        # Correct creation
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')

        # Service 1
        p11 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p11['id'], 'h1')
        p12 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p12['id'], 'h1')
        pp1 = self.create_port_pair(ingress=p11['id'], egress=p12['id'],
                                    expected_res_status=201)['port_pair']
        # Service 2
        p21 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p21['id'], 'h2')
        p22 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p22['id'], 'h2')
        pp2 = self.create_port_pair(ingress=p21['id'], egress=p22['id'],
                                    expected_res_status=201)['port_pair']
        # This goes through
        ppg1 = self.create_port_pair_group(
            port_pairs=[pp1['id'], pp2['id']],
            expected_res_status=201)['port_pair_group']
        # Use invalid pairs
        net3 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net3, '192.168.0.1', '192.168.0.0/24')
        p31 = self._make_port(self.fmt, net3['network']['id'])['port']
        self._bind_port_to_host(p31['id'], 'h1')
        pp3 = self.create_port_pair(ingress=p21['id'], egress=p31['id'],
                                    expected_res_status=201)['port_pair']
        self.delete_port_pair_group(ppg1['id'])
        self.create_port_pair_group(port_pairs=[pp1['id'], pp3['id']],
                                    expected_res_status=500)
        # Works with only one PP
        ppg2 = self.create_port_pair_group(
            port_pairs=[pp3['id']],
            expected_res_status=201)['port_pair_group']
        # But update fails
        self.update_port_pair_group(
            ppg2['id'], port_pairs=[pp3['id'], pp1['id']],
            expected_res_status=500)

    def test_ppg_update(self):
        # Correct creation
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')

        # Service 1
        p11 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p11['id'], 'h1')
        p12 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p12['id'], 'h1')
        pp1 = self.create_port_pair(ingress=p11['id'], egress=p12['id'],
                                    expected_res_status=201)['port_pair']
        # This goes through
        ppg1 = self.create_port_pair_group(
            port_pairs=[pp1['id']],
            expected_res_status=201)['port_pair_group']
        # Same ID update works
        self.update_port_pair_group(ppg1['id'], port_pairs=[pp1['id']],
                                    expected_res_status=200)

    def test_healthcheck_group(self):
        # Correct creation
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')

        # Service 1
        p11 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p11['id'], 'h1')
        p12 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p12['id'], 'h1')
        pp1 = self.create_port_pair(ingress=p11['id'], egress=p12['id'],
                                    expected_res_status=201)['port_pair']
        ppg1 = self.create_port_pair_group(
            port_pairs=[pp1['id']], port_pair_group_parameters={
                'healthcheck_type': 'tcp', 'healthcheck_frequency': 60,
                'healthcheck_tcp_port': 8080},
            expected_res_status=201)['port_pair_group']
        self.assertEqual('tcp', ppg1['port_pair_group_parameters'][
            'healthcheck_type'])
        self.assertEqual(60, ppg1['port_pair_group_parameters'][
            'healthcheck_frequency'])
        self.assertEqual(8080, ppg1['port_pair_group_parameters'][
            'healthcheck_tcp_port'])
        self.delete_port_pair_group(ppg1['id'])
        self.create_port_pair_group(
            port_pairs=[pp1['id']], port_pair_group_parameters={
                'healthcheck_type': 'no', 'healthcheck_frequency': 60,
                'healthcheck_tcp_port': 8080},
            expected_res_status=400)
        self.create_port_pair_group(
            port_pairs=[pp1['id']], port_pair_group_parameters={
                'healthcheck_type': 'tcp', 'healthcheck_frequency': -1,
                'healthcheck_tcp_port': 8080},
            expected_res_status=400)
        self.create_port_pair_group(
            port_pairs=[pp1['id']], port_pair_group_parameters={
                'healthcheck_type': 'tcp', 'healthcheck_frequency': 60,
                'healthcheck_tcp_port': 80800},
            expected_res_status=400)
        ppg1 = self.create_port_pair_group(
            port_pairs=[pp1['id']], port_pair_group_parameters={
                'healthcheck_type': 'icmp'},
            expected_res_status=201)['port_pair_group']
        self.assertEqual('icmp', ppg1['port_pair_group_parameters'][
            'healthcheck_type'])
        self.assertTrue('check_frequency' not in ppg1[
            'port_pair_group_parameters'])
        self.assertTrue('tcp_port' not in ppg1[
            'port_pair_group_parameters'])


class TestFlowClassifier(TestAIMServiceFunctionChainingBase):

    def test_fc_validation(self):
        # Correct classifier
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')

        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net1, '192.168.1.1', '192.168.1.0/24')
        fc = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=201)['flow_classifier']
        self.delete_flow_classifier(fc['id'], expected_res_status=204)
        # Wrong FCs
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id']},
            source_ip_prefix='192.168.0.0/24', expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id']},
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={'logical_source_network': net1['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24', expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={
                'logical_destination_network': net2['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24', expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net1['network']['id']},
            source_ip_prefix='192.168.0.0/24', expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': ''},
            source_ip_prefix='192.168.0.0/24', expected_res_status=400)
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id'],
            }, source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.0.0/24', expected_res_status=400)
        self._delete_network(net2['network']['id'])
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net1['network']['id'],
                'logical_destination_network': net2['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24', expected_res_status=404)
        net_svi = self._make_network(self.fmt, 'net_svi', True,
                                     arg_list=self.extension_attributes,
                                     **{'apic:svi': True})
        self._make_subnet(self.fmt, net_svi, '192.168.0.1', '192.168.0.0/24')
        fc = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net_svi['network']['id'],
                'logical_destination_network': net_svi['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=201)['flow_classifier']

        # Same subnets, different networks.
        net3 = self._make_network(self.fmt, 'net3', True)
        self._make_subnet(self.fmt, net1, '192.168.2.1', '192.168.2.0/24')
        net4 = self._make_network(self.fmt, 'net4', True)
        self._make_subnet(self.fmt, net1, '192.168.3.1', '192.168.3.0/24')
        self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net3['network']['id'],
                'logical_destination_network': net4['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=201)

        self.delete_flow_classifier(fc['id'], expected_res_status=204)


class TestPortChain(TestAIMServiceFunctionChainingBase):

    def setUp(self, *args, **kwargs):
        super(TestPortChain, self).setUp()
        self.src_svi = False
        self.dst_svi = False

    def _get_port_network(self, port_id):
        port = self._show_port(port_id)
        return self._show_network(port['network_id'])

    def test_pc_validation(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        ppg2 = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        # Same classifier is not allowed.
        self.create_port_chain(port_pair_groups=[ppg2['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=409)
        self.update_port_chain(
            pc['id'], port_pair_groups=[ppg['id'], ppg2['id']],
            expected_res_status=200)

    def test_pc_validation_network_conflict(self):
        nets = []
        for i in range(3):
            net = self._make_network(self.fmt, 'net1', True)
            self._make_subnet(self.fmt, net, '192.168.%s.1' % i,
                              '192.168.%s.0/24' % i)
            nets.append(net['network']['id'])
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2, leftn_id=nets[0],
                                      rightn_id=nets[1])
        ppg2 = self._create_simple_ppg(pairs=2, leftn_id=nets[0],
                                       rightn_id=nets[1])
        # Conflict with only one network
        ppg3 = self._create_simple_ppg(pairs=2, leftn_id=nets[0],
                                       rightn_id=nets[2])
        self.create_port_chain(port_pair_groups=[ppg['id'], ppg2['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=500)
        self.create_port_chain(port_pair_groups=[ppg['id'], ppg3['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=500)

    def test_pc_mapping(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_pc_mapping_two_flowcs(self):
        pc = self._create_simple_port_chain(
            flowcs=2, ppgs=3, ppg_args=[{'pairs': 1}, {'pairs': 2},
                                        {'pairs': 3}])
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_pc_mapping_no_flowcs(self):
        pc = self._create_simple_port_chain(
            flowcs=0, ppgs=3, ppg_args=[{'pairs': 1}, {'pairs': 2},
                                        {'pairs': 3}])
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_ppg_update(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        pps = ppg['port_pairs']
        # remove one pp
        ppg = self.update_port_pair_group(
            ppg['id'], port_pairs=[pps[0]])['port_pair_group']
        self._verify_pc_mapping(pc)
        # Replace pp
        ppg = self.update_port_pair_group(
            ppg['id'], port_pairs=[pps[1]])['port_pair_group']
        self._verify_pc_mapping(pc)
        # Add pp
        ppg = self.update_port_pair_group(
            ppg['id'], port_pairs=pps)['port_pair_group']
        self._verify_pc_mapping(pc)

    def test_flowc_update(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=1)
        self.create_port_chain(port_pair_groups=[ppg['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=201)

        res = self._delete_network(
            fc['l7_parameters']['logical_source_network'])
        self.assertTrue(res.status_int >= 400)
        res = self._delete_network(
            fc['l7_parameters']['logical_destination_network'])
        self.assertTrue(res.status_int >= 400)

    def test_vrf_update(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        self.create_port_chain(port_pair_groups=[ppg['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=201)
        self.aim_mgr.create(
            self._aim_context, aim_res.EndpointGroup(
                tenant_name='new', app_profile_name='new', name='new'))
        net_db = self._plugin._get_network(
            self._ctx, fc['l7_parameters']['logical_source_network'])
        self.assertRaises(c_exc.CallbackFailure,
                          self.aim_mech._set_network_vrf_and_notify,
                          self._ctx, net_db.aim_mapping,
                          aim_res.VRF(tenant_name='new', name='new'))
        net_db = self._plugin._get_network(
            self._ctx, fc['l7_parameters']['logical_destination_network'])
        self.assertRaises(c_exc.CallbackFailure,
                          self.aim_mech._set_network_vrf_and_notify,
                          self._ctx, net_db.aim_mapping,
                          aim_res.VRF(tenant_name='new', name='new'))
        # Also changing EPG affects PC if tenant changes
        self.assertRaises(c_exc.CallbackFailure,
                          self.aim_mech._set_network_epg_and_notify,
                          self._ctx, net_db.aim_mapping,
                          aim_res.EndpointGroup(tenant_name='new',
                                                app_profile_name='new',
                                                name='new'))
        # But it doesn't if tenant stays the same
        if not self.dst_svi:
            net_db = self._plugin._get_network(
                self._ctx,
                fc['l7_parameters']['logical_destination_network'])
            self.aim_mgr.create(
                self._aim_context, aim_res.EndpointGroup(
                    tenant_name=net_db.aim_mapping.epg_tenant_name,
                    app_profile_name='new', name='new'))
            self.aim_mech._set_network_epg_and_notify(
                self._ctx, net_db.aim_mapping, aim_res.EndpointGroup(
                    tenant_name=net_db.aim_mapping.epg_tenant_name,
                    app_profile_name='new', name='new'))

        pp = self.show_port_pair(ppg['port_pairs'][0])['port_pair']
        net = self._get_port_network(pp['ingress'])
        net_db = self._plugin._get_network(self._ctx, net['id'])
        # Modifying EPG in service nets has no effect
        self.aim_mech._set_network_epg_and_notify(
            self._ctx, net_db.aim_mapping,
            aim_res.EndpointGroup(tenant_name='new', app_profile_name='new',
                                  name='new'))
        # But it fails when VRF is changed
        self.assertRaises(c_exc.CallbackFailure,
                          self.aim_mech._set_network_vrf_and_notify,
                          self._ctx, net_db.aim_mapping,
                          aim_res.VRF(tenant_name='new', name='new'))

    def test_pc_mapping_no_host_mapping(self):
        ctx = self._aim_context
        self.aim_mgr.delete_all(ctx, aim_infra.HostDomainMappingV2)
        # Since one physdom exists, everything works just fine
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)
        # If I also delete the physdom, everything fails
        self.aim_mgr.delete(ctx, self.physdom)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=500)

    def test_pc_mapping_same_provider_diff_consumer(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        # New classifier with only one change in subnet
        fc2 = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': fc[
                    'l7_parameters']['logical_source_network'],
                'logical_destination_network': fc[
                    'l7_parameters']['logical_destination_network']},
            source_ip_prefix='192.168.3.0/24',
            destination_ip_prefix=fc['destination_ip_prefix'],
            expected_res_status=201)['flow_classifier']

        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        pc = self.update_port_chain(pc['id'],
                                    flow_classifiers=[fc['id'], fc2['id']],
                                    expected_res_status=200)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_pc_mapping_default_sub_dst(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        # New classifier with only one change in subnet
        fc2 = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': fc[
                    'l7_parameters']['logical_source_network'],
                'logical_destination_network': fc[
                    'l7_parameters']['logical_destination_network']},
            source_ip_prefix=fc['source_ip_prefix'],
            destination_ip_prefix='0.0.0.0/0',
            expected_res_status=201)['flow_classifier']

        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc2['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_pc_mapping_default_sub_src(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        # New classifier with only one change in subnet
        fc2 = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': fc[
                    'l7_parameters']['logical_source_network'],
                'logical_destination_network': fc[
                    'l7_parameters']['logical_destination_network']},
            source_ip_prefix='0.0.0.0/0',
            destination_ip_prefix=fc['destination_ip_prefix'],
            expected_res_status=201)['flow_classifier']

        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc2['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_port_pair_device_migration(self):

        def verify_port_in_host(port, host):
            dci = self.aim_mgr.find(
                self._aim_context, aim_sg.ConcreteDeviceInterface,
                name='prt_' + port['id'])[0]
            self.assertEqual(self.path_by_host[host], dci.path)

        ppg = self._create_simple_ppg(pairs=1)
        pp = self.show_port_pair(ppg['port_pairs'][0])['port_pair']
        iprt = self._show_port(pp['ingress'])
        eprt = self._show_port(pp['egress'])
        # Ports are initially bound to H1
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        # Rebind completely first port, then second
        iprt = self._unbind_port(iprt['id'])['port']
        verify_port_in_host(iprt, 'h1')
        self._plugin.update_port_status(self._ctx, iprt['id'], 'BUILD')
        verify_port_in_host(iprt, 'h1')
        self._bind_port_to_host(iprt['id'], 'h2')
        verify_port_in_host(iprt, 'h1')
        self._plugin.update_port_status(self._ctx, iprt['id'], 'ACTIVE')
        verify_port_in_host(iprt, 'h2')

        eprt = self._unbind_port(eprt['id'])['port']
        verify_port_in_host(eprt, 'h1')
        self._plugin.update_port_status(self._ctx, eprt['id'], 'BUILD')
        verify_port_in_host(eprt, 'h1')
        self._bind_port_to_host(eprt['id'], 'h2')
        verify_port_in_host(eprt, 'h1')
        self._plugin.update_port_status(self._ctx, iprt['id'], 'BUILD')
        self._plugin.update_port_status(self._ctx, eprt['id'], 'ACTIVE')
        # Other is not active
        verify_port_in_host(eprt, 'h1')
        self._plugin.update_port_status(self._ctx, iprt['id'], 'ACTIVE')
        verify_port_in_host(eprt, 'h2')
        self._verify_pc_mapping(pc)

    # Enable once fixed on the SVI side.
    def _test_pc_mapping_default_sub_ipv6(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        # New classifier with only one change in subnet
        fc2 = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': fc[
                    'l7_parameters']['logical_source_network'],
                'logical_destination_network': fc[
                    'l7_parameters']['logical_destination_network']},
            source_ip_prefix='::/0', destination_ip_prefix='::/0',
            expected_res_status=201, ethertype='IPv6')['flow_classifier']

        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc2['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_pc_max_ppg_validation(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg1 = self._create_simple_ppg(pairs=1)
        ppg2 = self._create_simple_ppg(pairs=1)
        ppg3 = self._create_simple_ppg(pairs=1)
        ppg4 = self._create_simple_ppg(pairs=1)
        self.create_port_chain(port_pair_groups=[ppg1['id'], ppg2['id'],
                                                 ppg3['id'], ppg4['id']],
                               flow_classifiers=[fc['id']],
                               expected_res_status=400)
        pc = self.create_port_chain(port_pair_groups=[ppg1['id'],
                                                      ppg2['id'],
                                                      ppg3['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self.update_port_chain(pc['id'],
                               port_pair_groups=[ppg1['id'], ppg2['id'],
                                                 ppg3['id'], ppg4['id']],
                               expected_res_status=500)

    def test_pc_move_fc(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        fcs = [fc]
        for i in range(3):
            fcs.append(self.create_flow_classifier(
                l7_parameters={
                    'logical_source_network': fc[
                        'l7_parameters']['logical_source_network'],
                    'logical_destination_network': fc[
                        'l7_parameters']['logical_destination_network']},
                source_ip_prefix='192.198.%s.0/24' % (i + 3),
                destination_ip_prefix=fc['destination_ip_prefix'],
                expected_res_status=201)['flow_classifier'])
        # We have four FCs
        ppg1 = self._create_simple_ppg(pairs=1)
        ppg2 = self._create_simple_ppg(pairs=1)
        pc1 = self.create_port_chain(port_pair_groups=[ppg1['id']],
                                     flow_classifiers=[fcs[0]['id'],
                                                       fcs[1]['id']],
                                     expected_res_status=201)['port_chain']
        pc2 = self.create_port_chain(port_pair_groups=[ppg2['id']],
                                     flow_classifiers=[fcs[2]['id'],
                                                       fcs[3]['id']],
                                     expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc1, multiple=True)
        self._verify_pc_mapping(pc2, multiple=True)
        # Remove FC 2
        pc1 = self.update_port_chain(pc1['id'],
                                     flow_classifiers=[fcs[0]['id']],
                                     expected_res_status=200)['port_chain']
        self._verify_pc_mapping(pc1, multiple=True)
        self._verify_pc_mapping(pc2, multiple=True)
        if self.dst_svi:
            dst_net_id = fc['l7_parameters']['logical_destination_network']
            ext_net = self.aim_mgr.find(
                self._aim_context, aim_res.ExternalNetwork,
                name=fc['destination_ip_prefix'].replace(
                    '/', '_') + '_' + 'net_' + dst_net_id)[0]
            self.assertEqual(2, len(ext_net.provided_contract_names))
            self.delete_port_chain(pc1['id'])
            ext_net = self.aim_mgr.get(self._aim_context, ext_net)
            self.assertEqual(1, len(ext_net.provided_contract_names))
            self.delete_port_chain(pc2['id'])
            self.assertIsNone(self.aim_mgr.get(self._aim_context, ext_net))

    def test_delete_no_contract(self):
        pc = self._create_simple_port_chain()
        epgs = self.aim_mgr.find(self._aim_context, aim_res.EndpointGroup)
        extn = self.aim_mgr.find(self._aim_context, aim_res.ExternalNetwork)
        for res in epgs + extn:
            self.aim_mgr.update(self._aim_context, res,
                                consumed_contract_names=[])
            self.aim_mgr.update(self._aim_context, res,
                                provided_contract_names=[])
        self._verify_pc_delete(pc)

    def test_port_link_update(self):
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')
        p1 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p1['id'], 'h1')

        p2 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p2['id'], 'h2')

        pp = self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                                   expected_res_status=201)['port_pair']
        ppg = self.create_port_pair_group(
            port_pairs=[pp['id']], expected_res_status=201)['port_pair_group']
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self.path_by_host['h2'] = 'topology/pod-1/paths-103/pathep-[eth3/1]'
        self.aim_mech.update_link(
            self._context, 'h2', 'eth0', 'aa:bb', '103',
            '3', '1',
            port_description='topology/pod-1/paths-103/pathep-[eth3/1]')
        self._verify_pc_mapping(pc)
        # Test Removal
        self.path_by_host.pop('h2')
        self.aim_mech.delete_link(self._context, 'h2', 'eth0', 'aa:bb', '103',
                                  '3', '1')
        self._verify_pc_mapping(pc)

    def test_port_no_host_link(self):
        net1 = self._make_network(self.fmt, 'net1', True)
        self._make_subnet(self.fmt, net1, '192.168.0.1', '192.168.0.0/24')
        net2 = self._make_network(self.fmt, 'net2', True)
        self._make_subnet(self.fmt, net2, '192.168.1.1', '192.168.1.0/24')
        p1 = self._make_port(self.fmt, net1['network']['id'])['port']
        self._bind_port_to_host(p1['id'], 'h-nolink')

        p2 = self._make_port(self.fmt, net2['network']['id'])['port']
        self._bind_port_to_host(p2['id'], 'h-nolink')

        pp = self.create_port_pair(ingress=p1['id'], egress=p2['id'],
                                   expected_res_status=201)['port_pair']
        ppg = self.create_port_pair_group(
            port_pairs=[pp['id']], expected_res_status=201)['port_pair_group']
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)

    def test_pc_update_flowc(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self.update_flow_classifier(
            fc['id'], name='new_name',
            expected_res_status=200)
        self._verify_pc_mapping(pc)
        self.update_flow_classifier(fc['id'], name='newname')
        self._verify_pc_mapping(pc)

    def test_pc_validation_hcheck(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        ppg = self._create_simple_ppg(pairs=2, check_type='icmp')
        ppg2 = self._create_simple_ppg(pairs=2, check_type='tcp',
                                       check_freq=31, check_port=90)
        pc = self.create_port_chain(port_pair_groups=[ppg['id'], ppg2['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)

    def test_same_provider_subnet(self):
        fc = self._create_simple_flowc(src_svi=self.src_svi,
                                       dst_svi=self.dst_svi)
        fcs = [fc]
        for i in range(3):
            fcs.append(self.create_flow_classifier(
                l7_parameters=fc['l7_parameters'],
                source_ip_prefix='192.198.%s.0/24' % (i + 3),
                destination_ip_prefix=fc['destination_ip_prefix'],
                expected_res_status=201)['flow_classifier'])
        # We have four FCs
        ppg1 = self._create_simple_ppg(pairs=1)
        self.create_port_chain(port_pair_groups=[ppg1['id']],
                               flow_classifiers=[fc['id'] for fc in fcs],
                               expected_res_status=201)


class TestPortChainSVI(TestPortChain):

    def setUp(self, *args, **kwargs):
        super(TestPortChainSVI, self).setUp()
        self.src_svi = True
        self.dst_svi = True

    def test_pc_flowc_same_network(self):
        # Need to make sure the SVI network is in the same common
        # unrouted vrf as other service chain resources do.
        router = self._make_router(
            self.fmt, self._tenant_id, 'router1')['router']
        vrf_name = 'openstack_UnroutedVRF'
        vrf = aim_res.VRF(tenant_name='common', name=vrf_name,
                          monitored=True)
        as1 = self._make_address_scope_for_vrf(
            vrf.dn, name='as1')['address_scope']
        pool = self._make_subnetpool(
            self.fmt, ['192.168.0.0/8'], name='sp',
            address_scope_id=as1['id'], tenant_id=as1['tenant_id'],
            default_prefixlen=24)['subnetpool']
        net_svi = self._make_network(self.fmt, 'net_svi', True,
                                     arg_list=self.extension_attributes,
                                     **{'apic:svi': True})
        subnet1 = self._make_subnet(
            self.fmt, net_svi, '192.168.0.1', '192.168.0.0/24',
            subnetpool_id=pool['id'])['subnet']
        self.l3_plugin.add_router_interface(
            context.get_admin_context(), router['id'],
            {'subnet_id': subnet1['id']})
        fc = self.create_flow_classifier(
            l7_parameters={
                'logical_source_network': net_svi['network']['id'],
                'logical_destination_network': net_svi['network']['id']},
            source_ip_prefix='192.168.0.0/24',
            destination_ip_prefix='192.168.1.0/24',
            expected_res_status=201)['flow_classifier']
        ppg = self._create_simple_ppg(pairs=2)
        pc = self.create_port_chain(port_pair_groups=[ppg['id']],
                                    flow_classifiers=[fc['id']],
                                    expected_res_status=201)['port_chain']
        self._verify_pc_mapping(pc)
        self._verify_pc_delete(pc)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""apic_per-tenant-nat-epg

Revision ID: 98862f2d814e
Revises: 12c1bc8d7026
Create Date: 2016-06-22 17:36:28.386526

"""

# revision identifiers, used by Alembic.
revision = '98862f2d814e'
down_revision = '12c1bc8d7026'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'gp_apic_tenant_specific_nat_epg',
        sa.Column('external_segment_id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('external_segment_id', 'tenant_id'),
        sa.ForeignKeyConstraint(['external_segment_id'],
                                ['gp_external_segments.id'],
                                ondelete='CASCADE',
                                name='gp_apic_ptne_fk_es'))


def downgrade():
    op.drop_table('gp_apic_tenant_specific_nat_epg')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.common import config  # noqa
from neutron import context as n_context
from neutron.plugins.common import constants as pconst
from oslo_config import cfg

from gbpservice.neutron.services.servicechain.plugins.ncp import model
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_gp_driver)
from gbpservice.neutron.tests.unit.services.servicechain.ncp import (
    test_ncp_plugin as base)


class TrafficStitchingPlumberTestCase(base.NodeCompositionPluginTestCase):

    def setUp(self):
        cfg.CONF.set_override('policy_drivers', ['implicit_policy',
                                                 'resource_mapping'],
                              group='group_policy')
        cfg.CONF.set_override('allow_overlapping_ips', True)
        cfg.CONF.set_override(
            'extension_drivers', ['proxy_group'], group='group_policy')
        super(TrafficStitchingPlumberTestCase, self).setUp(
            node_drivers=['node_dummy'], node_plumber='stitching_plumber',
            core_plugin=test_gp_driver.CORE_PLUGIN)
        res = mock.patch('neutron.db.l3_db.L3_NAT_dbonly_mixin.'
                         '_check_router_needs_rescheduling').start()
        res.return_value = None
        self.driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj
        self.driver.get_plumbing_info = mock.Mock()
        self.driver.get_plumbing_info.return_value = {}

    def test_one_gateway_pt_prov_cons(self):
        context = n_context.get_admin_context()
        self.driver.get_plumbing_info.return_value = {
            'provider': [{}], 'consumer': [{}], 'plumbing_type': 'gateway'}
        provider, consumer, node = self._create_simple_chain()
        provider = self.show_policy_target_group(
            provider['id'])['policy_target_group']
        # Verify Service PT created and correctly placed
        targets = model.get_service_targets(context.session)
        self.assertEqual(2, len(targets))
        old_relationship = None
        for target in targets:
            self.assertEqual(node['id'], target.servicechain_node_id)
            pt = self.show_policy_target(
                target.policy_target_id)['policy_target']
            if target.relationship == 'provider':
                self.assertEqual(provider['id'],
                                 pt['policy_target_group_id'])
                self.assertTrue(pt['group_default_gateway'])
                self.assertFalse(pt['proxy_gateway'])
            else:
                # Consumer side a proxy group exists
                self.assertEqual(provider['proxy_group_id'],
                                 pt['policy_target_group_id'])
                self.assertFalse(pt['group_default_gateway'])
                self.assertTrue(pt['proxy_gateway'])

            self.assertNotEqual(old_relationship, target.relationship)
            old_relationship = target.relationship
            port = self._get_object('ports', pt['port_id'], self.api)['port']
            self.assertTrue(port['name'].startswith('pt_service_target_'),
                            "Port name doesn't start with 'pt_service_target_"
                            "'.\nport:\n%s\n" % port)

        self.update_policy_target_group(
            provider['id'], provided_policy_rule_sets={})
        # With chain deletion, also the Service PTs are deleted
        new_targets = model.get_service_targets(context.session)
        self.assertEqual(0, len(new_targets))
        for target in targets:
            self.show_policy_target(
                target.policy_target_id, expected_res_status=404)
        provider = self.show_policy_target_group(
            provider['id'])['policy_target_group']
        self.assertIsNone(provider['proxy_group_id'])

    def test_multiple_endpoint_pt_provider(self):
        context = n_context.get_admin_context()
        self.driver.get_plumbing_info.return_value = {
            'provider': [{}, {}], 'consumer': [], 'plumbing_type': 'endpoint'}
        provider, consumer, node = self._create_simple_chain()
        provider = self.show_policy_target_group(
            provider['id'])['policy_target_group']
        # Verify Service PT created and contains proper name, description
        targets = model.get_service_targets(context.session)
        self.assertEqual(2, len(targets))
        for target in targets:
            pt = self.show_policy_target(
                target.policy_target_id)['policy_target']
            self.assertEqual(provider['id'],
                             pt['policy_target_group_id'])
            self.assertTrue(pt['name'].startswith('tscp_endpoint_service'),
                            "Policy Target name doesn't start with "
                            "'tscp_endpoint_service'.\npt:\n%s\n" % pt)
            self.assertTrue(node['id'] in pt['description'],
                            "Policy Target description doesn't contains "
                            " node id.\nnode:\n%s\n" % node)

            port = self._get_object('ports', pt['port_id'], self.api)['port']
            self.assertTrue(port['name'].startswith(
                            'pt_tscp_endpoint_service'),
                            "Port name doesn't start with "
                            "'pt_tscp_endpoint_service'.\nport:\n%s\n" % port)

        self.update_policy_target_group(
            provider['id'], provided_policy_rule_sets={})
        # With chain deletion, also the Service PTs are deleted
        new_targets = model.get_service_targets(context.session)
        self.assertEqual(0, len(new_targets))
        for target in targets:
            self.show_policy_target(
                target.policy_target_id, expected_res_status=404)
        provider = self.show_policy_target_group(
            provider['id'])['policy_target_group']
        self.assertIsNone(provider['proxy_group_id'])

    def get_plumbing_info_base(self, context):
        service_type = context.current_profile['service_type']
        plumbing_request = {'management': [], 'provider': [{}],
                            'consumer': [{}]}

        if service_type in [pconst.FIREWALL]:
            plumbing_request['plumbing_type'] = 'gateway'
        else:
            plumbing_request = {}
        return plumbing_request

    def test_get_service_targets_in_chain(self):
        context = n_context.get_admin_context()
        self.driver.get_plumbing_info = self.get_plumbing_info_base
        lb_prof = self._create_service_profile(
            service_type='LOADBALANCERV2',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        lb_node = self.create_servicechain_node(
            service_profile_id=lb_prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        fw_prof = self._create_service_profile(
            service_type='FIREWALL',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        fw_node = self.create_servicechain_node(
            service_profile_id=fw_prof['id'],
            config='{}')['servicechain_node']

        self._create_chain_with_nodes([fw_node['id'], lb_node['id']])

        targets = model.get_service_targets(context.session)
        self.assertEqual(2, len(targets))

    def test_ptg_delete(self):
        self.driver.get_plumbing_info.return_value = {
            'provider': [{}], 'consumer': [{}],
            'plumbing_type': 'transparent'}
        provider, _, _ = self._create_simple_service_chain()
        # Deleting a PTG will fail because of existing PTs
        self.delete_policy_target_group(provider['id'],
                                        expected_res_status=204)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.api.v2 import attributes
from neutron_lib.api import converters as conv
from neutron_lib import constants


EXTENDED_ATTRIBUTES_2_0 = {
    attributes.SUBNETPOOLS: {
        'is_implicit': {'allow_post': True, 'allow_put': True,
                        'default': False,
                        'convert_to': conv.convert_to_boolean,
                        'is_visible': True, 'required_by_policy': True,
                        'enforce_policy': True},
    },
}


class Implicit_subnetpools(extensions.ExtensionDescriptor):
    """Extension class supporting default subnetpools."""

    @classmethod
    def get_name(cls):
        return "Implicit Subnetpools"

    @classmethod
    def get_alias(cls):
        return "implicit-subnetpools"

    @classmethod
    def get_description(cls):
        return "Provides ability to mark a subnetpool as implicit"

    @classmethod
    def get_updated(cls):
        return "2017-01-11T18:00:00-00:00"

    def get_required_extensions(self):
        return [constants.SUBNET_ALLOCATION_EXT_ALIAS]

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.api.v2 import attributes
from neutron.extensions import address_scope
from neutron_lib.api import converters as conv

ALIAS = 'cisco-apic'

DIST_NAMES = 'apic:distinguished_names'
SYNC_STATE = 'apic:synchronization_state'
NAT_TYPE = 'apic:nat_type'
SNAT_HOST_POOL = 'apic:snat_host_pool'
EXTERNAL_CIDRS = 'apic:external_cidrs'
SVI = 'apic:svi'
BGP = 'apic:bgp_enable'
BGP_ASN = 'apic:bgp_asn'
BGP_TYPE = 'apic:bgp_type'

BD = 'BridgeDomain'
EPG = 'EndpointGroup'
SUBNET = 'Subnet'
VRF = 'VRF'
EXTERNAL_NETWORK = 'ExternalNetwork'
AP = 'ApplicationProfile'

SYNC_SYNCED = 'synced'
SYNC_BUILD = 'build'
SYNC_ERROR = 'error'
SYNC_NOT_APPLICABLE = 'N/A'

APIC_ATTRIBUTES = {
    DIST_NAMES: {'allow_post': False, 'allow_put': False, 'is_visible': True},
    SYNC_STATE: {'allow_post': False, 'allow_put': False, 'is_visible': True}
}

NET_ATTRIBUTES = {
    SVI: {
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    BGP: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    BGP_TYPE: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': 'default_export',
        'validate': {'type:values': ['default_export', '']},
    },
    BGP_ASN: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': "0",
        'validate': {'type:non_negative': None},
    },
}

EXT_NET_ATTRIBUTES = {
    DIST_NAMES: {
        # DN of corresponding APIC L3Out external network; can be
        # specified only on create.
        # Change 'allow_put' if updates on other DNs is allowed later,
        # and validate that ExternalNetwork DN may not be updated.
        'allow_post': True, 'allow_put': False,
        'is_visible': True,
        'default': None,
        'validate': {
            'type:dict_or_none': {
                EXTERNAL_NETWORK: {'type:string': None,
                                   'required': True}
            }
        }
    },
    NAT_TYPE: {
        # whether NAT is enabled, and if so its type
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': 'distributed',
        'validate': {'type:values': ['distributed', 'edge', '']},
    },
    EXTERNAL_CIDRS: {
        # Restrict external traffic to specified addresses
        'allow_put': True, 'allow_post': True,
        'is_visible': True, 'default': ['0.0.0.0/0'],
        'convert_to': conv.convert_none_to_empty_list,
        'validate': {'type:subnet_list': None},
    },
}

EXT_SUBNET_ATTRIBUTES = {
    SNAT_HOST_POOL: {
        # whether an external subnet should be used as a pool
        # for allocating host-based SNAT addresses
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    }
}

ADDRESS_SCOPE_ATTRIBUTES = {
    DIST_NAMES: {
        # DN of corresponding APIC VRF; can be specified only on create.
        # Change 'allow_put' if updates on other DNs is allowed later,
        # and validate that VRF DN may not be updated.
        'allow_post': True, 'allow_put': False,
        'is_visible': True,
        'default': None,
        'validate': {
            'type:dict_or_none': {
                VRF: {'type:string': None,
                      'required': True}
            }
        }
    }
}


EXTENDED_ATTRIBUTES_2_0 = {
    attributes.NETWORKS: dict(
        APIC_ATTRIBUTES.items() + EXT_NET_ATTRIBUTES.items() +
        NET_ATTRIBUTES.items()),
    attributes.SUBNETS: dict(
        APIC_ATTRIBUTES.items() + EXT_SUBNET_ATTRIBUTES.items()),
    address_scope.ADDRESS_SCOPES: dict(
        APIC_ATTRIBUTES.items() + ADDRESS_SCOPE_ATTRIBUTES.items())
}


class Cisco_apic(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Cisco APIC"

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return ("Extension exposing mapping of Neutron resources to Cisco "
                "APIC constructs")

    @classmethod
    def get_updated(cls):
        return "2016-03-31T12:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from heatclient import exc as heat_exc
from neutron.db import api as db_api
from neutron.db import models_v2 as ndb
from neutron.plugins.common import constants as pconst
from neutron_lib.db import model_base
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_serialization import jsonutils
import sqlalchemy as sa

from gbpservice._i18n import _LE
from gbpservice.neutron.services.servicechain.plugins.ncp import (
                                                    exceptions as exc)
from gbpservice.neutron.services.servicechain.plugins.ncp import driver_base
from gbpservice.neutron.services.servicechain.plugins.ncp.node_drivers import (
                                openstack_heat_api_client as heat_api_client)

LOG = logging.getLogger(__name__)

service_chain_opts = [
    cfg.IntOpt('stack_action_wait_time',
               default=15,
               help=_("Seconds to wait for pending stack operation "
                      "to complete")),
    cfg.IntOpt('delete_vip_port_retries',
               default=10,
               help=_("Retries to check if LB VIP port is deleted")),
    cfg.StrOpt('heat_uri',
               default='http://localhost:8004/v1',
               help=_("Heat API server address to instantiate services "
                      "specified in the service chain.")),
    cfg.StrOpt('exclude_pool_member_tag',
               default='ExcludePoolMember',
               help=_("Policy Targets created for the LB Pool Members should "
                      "have this tag in their description")),
]

cfg.CONF.register_opts(service_chain_opts, "heat_node_driver")
EXCLUDE_POOL_MEMBER_TAG = cfg.CONF.heat_node_driver.exclude_pool_member_tag
STACK_ACTION_WAIT_TIME = cfg.CONF.heat_node_driver.stack_action_wait_time
STACK_ACTION_RETRY_WAIT = 5  # Retry after every 5 seconds
DELETE_VIP_PORT_RETRIES = cfg.CONF.heat_node_driver.delete_vip_port_retries


class ServiceNodeInstanceStack(model_base.BASEV2):
    """ServiceChainInstance stacks owned by the Node driver."""

    __tablename__ = 'ncp_node_instance_stacks'
    sc_instance_id = sa.Column(sa.String(36),
                               nullable=False, primary_key=True)
    sc_node_id = sa.Column(sa.String(36),
                           nullable=False, primary_key=True)
    stack_id = sa.Column(sa.String(36),
                         nullable=False, primary_key=True)


class InvalidServiceType(exc.NodeCompositionPluginBadRequest):
    message = _("The Heat Node driver only supports the services "
                "Firewall and LB in a Service Chain")


class ServiceProfileRequired(exc.NodeCompositionPluginBadRequest):
    message = _("A Service profile is required in Service node")


class NodeVendorMismatch(exc.NodeCompositionPluginBadRequest):
    message = _("The Heat Node driver only handles nodes which have service "
                "profile with vendor name %(vendor)s")


class ServiceConfigNotJsonString(exc.NodeCompositionPluginBadRequest):
    message = _("Service config should be a json string for the Heat Node "
                "driver")


class HeatTemplateVersionNotSupported(exc.NodeCompositionPluginBadRequest):
    message = _("The Heat Node driver only supports AWS and HOT template "
                "formats for service node config")


class ServiceResourceDefinitionsMissing(exc.NodeCompositionPluginBadRequest):
    message = _("The Service template does not have service resources defined")


class HeatResourceMissing(exc.NodeCompositionPluginBadRequest):
    message = _("The service template requires the Resource %(resource)s for "
                "service type %(servicetype)s")


class ProfileUpdateNotSupported(exc.NodeCompositionPluginBadRequest):
    message = _("The Heat Node driver does not allow updating the "
                "service profile used by a Node")


class ServiceTypeUpdateNotSupported(exc.NodeCompositionPluginBadRequest):
    message = _("The Heat Node driver does not allow updating the "
                "service type used by a Node")


class HeatNodeDriver(driver_base.NodeDriverBase):

    vendor_name = 'heat_based_node_driver'
    initialized = False
    sc_supported_type = [pconst.LOADBALANCERV2, pconst.FIREWALL]
    required_heat_resources = {
        pconst.LOADBALANCERV2: ['OS::Neutron::LBaaS::LoadBalancer',
                                'OS::Neutron::LBaaS::Listener',
                                'OS::Neutron::LBaaS::Pool'],
        pconst.FIREWALL: ['OS::Neutron::Firewall',
                          'OS::Neutron::FirewallPolicy'],
    }

    @log.log_method_call
    def initialize(self, name):
        self.initialized = True
        self._name = name

    @log.log_method_call
    def get_plumbing_info(self, context):
        pass

    @log.log_method_call
    def validate_create(self, context):
        if context.current_profile is None:
            raise ServiceProfileRequired()
        if context.current_profile['vendor'].lower() != (
            self.vendor_name.lower()):
            raise NodeVendorMismatch(vendor=self.vendor_name)
        service_type = context.current_profile['service_type']
        if service_type not in self.sc_supported_type:
            raise InvalidServiceType()

        self._validate_service_config(context.current_node['config'],
                                      service_type)

    @log.log_method_call
    def validate_update(self, context):
        if not context.original_node:  # PT create/delete notifications
            return
        if context.current_profile != context.original_profile:
            raise ProfileUpdateNotSupported()
        if (context.current_node['service_type'] !=
            context.original_node['service_type']):
            raise ServiceTypeUpdateNotSupported()
        else:
            service_type = context.current_profile['service_type']
            self._validate_service_config(context.current_node['config'],
                                          service_type)

    def _validate_service_config(self, service_template, service_type):
        if not service_template:
            raise ServiceResourceDefinitionsMissing()

        try:
            service_template = jsonutils.loads(service_template)
        except Exception:
            raise ServiceConfigNotJsonString()

        if (not service_template.get('AWSTemplateFormatVersion') and
                not service_template.get('heat_template_version')):
                raise HeatTemplateVersionNotSupported()
        is_template_aws_version = service_template.get(
                                        'AWSTemplateFormatVersion', False)
        resources_key = 'Resources' if is_template_aws_version else 'resources'
        if not service_template.get(resources_key):
            raise ServiceResourceDefinitionsMissing()

        for resource_name in self.required_heat_resources[service_type]:
            param_key = self._get_heat_resource_key(
                            service_template[resources_key],
                            is_template_aws_version,
                            resource_name)
            if not param_key:
                raise HeatResourceMissing(resource=resource_name,
                                          servicetype=service_type)

    @log.log_method_call
    def create(self, context):
        heatclient = self._get_heat_client(context.plugin_context)

        stack_template, stack_params = self._fetch_template_and_params(context)

        stack_name = ("stack_" + context.instance['name'] +
                      context.current_node['name'] +
                      context.instance['id'][:8] +
                      context.current_node['id'][:8])
        # Heat does not accept space in stack name
        stack_name = stack_name.replace(" ", "")
        stack = heatclient.create(stack_name, stack_template, stack_params)

        self._insert_node_instance_stack_in_db(
            context.plugin_session, context.current_node['id'],
            context.instance['id'], stack['stack']['id'])

    @log.log_method_call
    def delete(self, context):
        stack_ids = self._get_node_instance_stacks(context.plugin_session,
                                                   context.current_node['id'],
                                                   context.instance['id'])
        heatclient = self._get_heat_client(context.plugin_context)

        for stack in stack_ids:
            vip_port_id = None
            try:
                rstr = heatclient.client.resources.get(stack_ids[0].stack_id,
                    'loadbalancer')
                vip_port_id = rstr.attributes['vip_port_id']
            except heat_exc.HTTPNotFound:
                # stack not found, so no need to process any further
                pass
            heatclient.delete(stack.stack_id)
            if vip_port_id:
                for x in range(0, DELETE_VIP_PORT_RETRIES):
                    # We intentionally get a new session so as to be
                    # able to read the updated DB
                    session = db_api.get_session()
                    vip_port = session.query(ndb.Port).filter_by(
                        id=vip_port_id).all()
                    if vip_port:
                        # heat stack delete is not finished yet, so try again
                        LOG.debug(("VIP port %s is not yet deleted"), vip_port)
                        LOG.debug(("Retry attempt; %s"), x + 1)
                        # Stack delete will at least take some minimal amount
                        # of time, hence we wait a little bit.
                        time.sleep(STACK_ACTION_WAIT_TIME)
                    else:
                        # we force a retry so that a new session can be
                        # used that will correctly reflect the VIP port as
                        # deleted and hence allow the subsequent policy driver
                        # to delete the VIP subnet
                        raise db_exc.RetryRequest(Exception)

        self._delete_node_instance_stack_in_db(context.plugin_session,
                                               context.current_node['id'],
                                               context.instance['id'])

    @log.log_method_call
    def update(self, context):
        heatclient = self._get_heat_client(context.plugin_context)

        stack_template, stack_params = self._fetch_template_and_params(context)

        stack_ids = self._get_node_instance_stacks(context.plugin_session,
                                                   context.current_node['id'],
                                                   context.instance['id'])
        for stack in stack_ids:
            self._wait_for_stack_operation_complete(
                                heatclient, stack.stack_id, 'update')
            heatclient.update(stack.stack_id, stack_template, stack_params)

    @log.log_method_call
    def update_policy_target_added(self, context, policy_target):
        if context.current_profile['service_type'] == pconst.LOADBALANCERV2:
            self.update(context)

    @log.log_method_call
    def update_policy_target_removed(self, context, policy_target):
        if context.current_profile['service_type'] == pconst.LOADBALANCERV2:
            self.update(context)

    @log.log_method_call
    def update_node_consumer_ptg_added(self, context, policy_target_group):
        pass

    @log.log_method_call
    def update_node_consumer_ptg_removed(self, context, policy_target_group):
        pass

    @log.log_method_call
    def notify_chain_parameters_updated(self, context):
        self.update(context)

    @log.log_method_call
    def policy_target_group_updated(self, context, old_policy_target_group,
                                    current_policy_target_group):
        pass

    def get_status(self, context):
        # TODO(Sumit): Needs to be implemented
        return {'status': '', 'status_details': ''}

    @property
    def name(self):
        return self._name

    def _get_heat_client(self, plugin_context):
        return heat_api_client.HeatClient(
                                plugin_context,
                                cfg.CONF.heat_node_driver.heat_uri)

    def _fetch_template_and_params(self, context):
        sc_instance = context.instance
        provider_ptg = context.provider
        # TODO(Magesh): Handle multiple subnets
        provider_ptg_subnet_id = provider_ptg['subnets'][0]
        consumer = context.consumer
        service_type = context.current_profile['service_type']

        stack_template = context.current_node.get('config')
        stack_template = jsonutils.loads(stack_template)
        config_param_values = sc_instance.get('config_param_values', {})
        stack_params = {}

        if config_param_values:
            config_param_values = jsonutils.loads(config_param_values)

        is_template_aws_version = stack_template.get(
                                        'AWSTemplateFormatVersion', False)

        if service_type == pconst.LOADBALANCERV2:
            self._generate_pool_members(context, stack_template,
                                        config_param_values,
                                        provider_ptg,
                                        is_template_aws_version)
        elif service_type == pconst.FIREWALL:
            provider_subnet = context.core_plugin.get_subnet(
                                context.plugin_context, provider_ptg_subnet_id)
            consumer_cidrs = []
            if consumer:
                if context.is_consumer_external:
                    # REVISIT(Magesh): Allowing the first destination which is
                    # 0/0 Validate and skip adding FW rule in case routes is
                    # not set
                    es = context.gbp_plugin.get_external_segment(
                        context.plugin_context,
                        consumer['external_segments'][0])
                    consumer_cidrs = [x['destination']
                                      for x in es['external_routes']]
                else:
                    consumer_subnet = context.core_plugin.get_subnet(
                        context._plugin_context, consumer['subnets'][0])
                    consumer_cidrs = [consumer_subnet['cidr']]
            provider_cidr = provider_subnet['cidr']
            self._update_template_with_firewall_rules(
                    context, provider_ptg, provider_cidr, consumer_cidrs,
                    stack_template, is_template_aws_version)

        node_params = (stack_template.get('Parameters')
                       or stack_template.get('parameters')
                       or [])
        for parameter in node_params:
            if parameter == "Subnet":
                stack_params[parameter] = provider_ptg_subnet_id
            elif parameter == "service_chain_metadata":
                stack_params[parameter] = sc_instance['id']
            elif parameter in config_param_values:
                stack_params[parameter] = config_param_values[parameter]
        return (stack_template, stack_params)

    def _wait_for_stack_operation_complete(self, heatclient, stack_id, action):
        time_waited = 0
        while True:
            try:
                stack = heatclient.get(stack_id)
                if stack.stack_status == 'DELETE_FAILED':
                    heatclient.delete(stack_id)
                elif stack.stack_status not in ['UPDATE_IN_PROGRESS',
                                                'DELETE_IN_PROGRESS']:
                    return
            except Exception:
                LOG.exception(_LE("Retrieving the stack %(stack)s failed."),
                              {'stack': stack_id})
                return
            else:
                time.sleep(STACK_ACTION_RETRY_WAIT)
                time_waited = time_waited + STACK_ACTION_RETRY_WAIT
                if time_waited >= STACK_ACTION_WAIT_TIME:
                    LOG.error(_LE("Stack %(action)s not completed within "
                                  "%(wait)s seconds"),
                              {'action': action,
                               'wait': STACK_ACTION_WAIT_TIME,
                               'stack': stack_id})
                    return

    def _delete_node_instance_stack_in_db(self, session, sc_node_id,
                                          sc_instance_id):
        with session.begin(subtransactions=True):
            stacks = (session.query(ServiceNodeInstanceStack).
                      filter_by(sc_node_id=sc_node_id).
                      filter_by(sc_instance_id=sc_instance_id).
                      all())
            for stack in stacks:
                session.delete(stack)

    def _insert_node_instance_stack_in_db(self, session, sc_node_id,
                                          sc_instance_id, stack_id):
        with session.begin(subtransactions=True):
            chainstack = ServiceNodeInstanceStack(
                sc_node_id=sc_node_id,
                sc_instance_id=sc_instance_id,
                stack_id=stack_id)
            session.add(chainstack)

    def _get_node_instance_stacks(self, session, sc_node_id=None,
                                  sc_instance_id=None):
        with session.begin(subtransactions=True):
            query = session.query(ServiceNodeInstanceStack)
            if sc_node_id:
                query = query.filter_by(sc_node_id=sc_node_id)
            if sc_instance_id:
                query = query.filter_by(sc_instance_id=sc_instance_id)
            return query.all()

    def _update_template_with_firewall_rules(self, context, provider_ptg,
                                             provider_cidr, consumer_cidrs,
                                             stack_template,
                                             is_template_aws_version):
        resources_key = ('Resources' if is_template_aws_version
                         else 'resources')
        properties_key = ('Properties' if is_template_aws_version
                          else 'properties')
        ref_key = 'Ref' if is_template_aws_version else 'get_resource'

        rule_num = 1
        rule_list = []
        for consumer_cidr in consumer_cidrs:
            rule_name = "Rule_" + str(rule_num)
            rule_num = rule_num + 1
            stack_template[resources_key][rule_name] = (
                self._generate_firewall_rule(
                    is_template_aws_version, context.classifier["protocol"],
                    context.classifier["port_range"],
                    provider_cidr, consumer_cidr))
            rule_list.append({ref_key: rule_name})

        resource_name = 'OS::Neutron::FirewallPolicy'
        fw_policy_key = self._get_heat_resource_key(
                            stack_template[resources_key],
                            is_template_aws_version,
                            resource_name)

        stack_template[resources_key][fw_policy_key][properties_key][
            'firewall_rules'] = rule_list

    def _generate_firewall_rule(self, is_template_aws_version, protocol,
                                destination_port, destination_cidr,
                                source_cidr):
        type_key = 'Type' if is_template_aws_version else 'type'
        properties_key = ('Properties' if is_template_aws_version
                          else 'properties')
        return {type_key: "OS::Neutron::FirewallRule",
                properties_key: {
                    "protocol": protocol,
                    "enabled": True,
                    "destination_port": destination_port,
                    "action": "allow",
                    "destination_ip_address": destination_cidr,
                    "source_ip_address": source_cidr}}

    def _generate_pool_members(self, context, stack_template,
                               config_param_values, provider_ptg,
                               is_template_aws_version):
        resources_key = 'Resources' if is_template_aws_version else 'resources'
        type_key = 'Type' if is_template_aws_version else 'type'
        member_ips = self._get_member_ips(context, provider_ptg)
        if not member_ips:
            return

        pool_res_name = None
        for resource in stack_template[resources_key]:
            if stack_template[resources_key][resource][type_key] == (
                    'OS::Neutron::LBaaS::Pool'):
                pool_res_name = resource
                break

        for member_ip in member_ips:
            member_name = 'mem-' + member_ip
            stack_template[resources_key][member_name] = (
                self._generate_pool_member_template(
                    context, is_template_aws_version,
                    pool_res_name, member_ip))

    def _generate_pool_member_template(self, context,
                                       is_template_aws_version,
                                       pool_res_name, member_ip):
        type_key = 'Type' if is_template_aws_version else 'type'
        properties_key = ('Properties' if is_template_aws_version
                          else 'properties')
        res_key = 'Ref' if is_template_aws_version else 'get_resource'
        return {type_key: "OS::Neutron::LBaaS::PoolMember",
                properties_key: {
                    "address": member_ip,
                    "admin_state_up": True,
                    "pool": {res_key: pool_res_name},
                    "protocol_port": {'get_param': 'app_port'},
                    "subnet": {'get_param': 'Subnet'},
                    "weight": 1}}

    def _get_member_ips(self, context, ptg):
        member_addresses = []
        policy_target_groups = context.gbp_plugin.get_policy_targets(
                context.plugin_context,
                filters={'id': ptg.get("policy_targets")})
        for policy_target in policy_target_groups:
            if EXCLUDE_POOL_MEMBER_TAG not in policy_target['description']:
                port_id = policy_target.get("port_id")
                if port_id:
                    port = context.core_plugin.get_port(
                                        context._plugin_context, port_id)
                    ip = port.get('fixed_ips')[0].get("ip_address")
                    member_addresses.append(ip)
        return member_addresses

    def _get_heat_resource_key(self, template_resource_dict,
                               is_template_aws_version, resource_name):
        type_key = 'Type' if is_template_aws_version else 'type'
        for key in template_resource_dict:
            if template_resource_dict[key].get(type_key) == resource_name:
                return key



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from argparse import Namespace

from gbpservice._i18n import _
from gbpservice.nfp.core import context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import threadpool as core_tp

LOG = nfp_logging.getLogger(__name__)


class InUse(Exception):

    """Exception raised when same task executor instance
        is fired twice or jobs
        added after executor is fired.
    """
    pass


def check_in_use(f):
    """Check if instance of task executor is already
        fired and executing jobs.
    """

    def wrapped(self, *args, **kwargs):
        if self.fired:
            raise InUse(_("Executor in use"))
        return f(self, *args, **kwargs)
    return wrapped


class TaskExecutor(object):

    """Executes given jobs in green threads.

        Any number of jobs can be added till executor
        is fired. When fired, executes all jobs in
        parallel in green threads. Waits for threads
        to complete, captures the return values of thread
        function.
        Caller can choose to pass result_store where the
        return value will be updated.
    """

    def __init__(self, jobs=0):
        if not jobs:
            self.thread_pool = core_tp.ThreadPool()
        else:
            self.thread_pool = core_tp.ThreadPool(thread_pool_size=jobs)

        self.pipe_line = []
        self.fired = False

    @check_in_use
    def add_job(self, id, func, *args, **kwargs):
        result_store = kwargs.pop('result_store', None)

        job = {
            'id': id, 'method': func,
            'args': args, 'kwargs': kwargs
        }

        if result_store is not None:
            job.update({'result_store': result_store})

        LOG.debug("TaskExecutor - (job - %s) added to pipeline",
                  (str(job)))

        self.pipe_line.append(job)

    def _complete(self):
        LOG.debug("TaskExecutor - complete")
        self.pipe_line = []
        self.fired = False

    def dispatch(self, job):
        context.init()
        return job['method'](*job['args'], **job['kwargs'])

    @check_in_use
    def fire(self):
        self.fired = True
        for job in self.pipe_line:
            LOG.debug(
                "TaskExecutor - (job - %s) dispatched",
                (str(job)))

            th = self.thread_pool.dispatch(self.dispatch, job)
            job['thread'] = th

        for job in self.pipe_line:
            result = job['thread'].wait()
            LOG.debug(
                "TaskExecutor - (job - %s) complete",
                (str(job)))

            job.pop('thread')
            job['result'] = result
            if 'result_store' in job.keys():
                job['result_store']['result'] = result

        done_jobs = self.pipe_line[:]
        self._complete()
        return done_jobs


class EventGraphExecutor(object):

    """Executor which executs a graph of events.

        An event graph can consist of events defined
        in any combination of parallel and sequence
        events. Executor will execute them in the
        order and manner specified.
        Eg., E1 -> (E2, E3)
                [E1 should execute after E2, E3 completes,
                 while E2 & E3 can happen in parallel]
            E2 -> (E4, E5)
                [E2 should execute after E4, E5 completes,
                 while E4 & E5 should happen in sequence]
            E3 -> (None)
                [No child events for E3]

        Executor will run the above graph and execute events
        in the exact specific order mentioned.
        At each level, parent event holds the result of child
        events, caller can use parent event complete notification
        to get the child events execution status.
    """

    def __init__(self, manager):
        self.manager = manager
        self.running = {}

    def add(self, graph):
        assert graph['id'] not in self.running.keys(), "Graph - %s \
            is already running" % (graph['id'])
        graph['results'] = dict.fromkeys(graph['data'])
        self.running[graph['id']] = graph
        self.run(graph['id'], graph['root'])

    def run(self, graph_id, node):
        graph = self.running[graph_id]
        leafs = self._leafs(graph['data'], node)
        if leafs == []:
            results = self._results(graph, node)
            self._schedule(node, results=results)
        else:
            self._dispatch(graph, leafs)

    def _results(self, graph, node):
        try:
            return self.running['results'][node]
        except KeyError:
            return []

    def _dispatch(self, graph, nodes):
        for node in nodes:
            event = self.manager.get_event(node)
            if event.sequence:
                self._schedule(node)
            else:
                self.run(graph['id'], node)

    def _leafs(self, tree, root):
        leafs = []
        try:
            leafs = tree[root]
        finally:
            return leafs

    def _root(self, graph, of):
        tree = graph['data']
        for root, nodes in six.iteritems(tree):
            if of in nodes:
                return root
        return None

    def _schedule(self, node, results=None):
        results = results or []
        event = self.manager.get_event(node)
        event.result = results
        self.manager._scheduled_new_event(event)

    def _graph(self, node):
        for graph in self.running.values():
            root = self._root(graph, node)
            if root:
                return graph

    def _prepare_result(self, node, result):
        result_obj = Namespace()
        key, id = node.split(':')
        result_obj.id = id
        result_obj.key = key
        result_obj.result = result
        return result_obj

    def _update_result(self, graph, root, result):
        if not graph['results'][root]:
            graph['results'][root] = []
        graph['results'][root].append(result)
        return graph['results'][root]

    def conntinue(self, completed_node, result):
        graph = self._graph(completed_node)
        if graph:
            if completed_node == graph['root']:
                # Graph is complete here, remove from running_instances
                self.running.pop(graph['id'])
            else:
                root = self._root(graph, completed_node)
                graph['data'][root].remove(completed_node)
                result = self._prepare_result(completed_node, result)
                results = self._update_result(graph, root, result)
                if graph['data'][root] == []:
                    self._schedule(root, results=results)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from heatclient import client as heat_client
from heatclient import exc as heat_exc

from gbpservice._i18n import _LW
from gbpservice.nfp.core import log as nfp_logging
LOG = nfp_logging.getLogger(__name__)

# We are overriding create and update for now because the upstream
# heat client class does not take timeout as argument


class HeatClient(object):

    def __init__(self, user_name, tenant, heat_uri, password=None,
                 auth_token=None, timeout_mins=30):
        api_version = "1"
        endpoint = "%s/%s" % (heat_uri, tenant)
        kwargs = {
            'token': auth_token,
            'username': user_name,
            'password': password
        }
        self.client = heat_client.Client(api_version, endpoint, **kwargs)
        self.stacks = self.client.stacks

        self.timeout_mins = timeout_mins
        # REVISIT(ashu): The base class is a old style class. We have to
        # change when it is updated
        # gbp_heat_api_client.HeatClient.__init__(
        #    self, context, heat_uri, password, auth_token)

    def create(self, name, data, parameters=None):
        fields = {
            'stack_name': name,
            'timeout_mins': self.timeout_mins,
            'disable_rollback': True,
            'password': data.get('password')
        }
        fields['template'] = data
        fields['parameters'] = parameters
        return self.stacks.create(**fields)

    def update(self, stack_id, data, parameters=None):
        fields = {
            'timeout_mins': self.timeout_mins,
            'password': data.get('password')
        }
        fields['template'] = data
        fields['parameters'] = parameters
        return self.stacks.update(stack_id, **fields)

    def delete(self, stack_id):
        try:
            self.stacks.delete(stack_id)
        except heat_exc.HTTPNotFound:
            LOG.warning(_LW("Stack %(stack)s created by service chain driver "
                            "is not found at cleanup"), {'stack': stack_id})

    def get(self, stack_id):
        return self.stacks.get(stack_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
import six

from gbpservice._i18n import _LI
from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.lib import data_filter
from gbpservice.contrib.nfp.configurator.lib import utils
from gbpservice.contrib.nfp.configurator.lib import vpn_constants as const
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)


class VpnaasRpcSender(data_filter.Filter):
    """
    Implements VPNaas response path to Neutron plugin.

    Methods of this class are invoked by the VPNaasEventHandler class
    for sending response from driver to the VPNaas Neutron plugin.

    """
    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, sc):
        self._sc = sc
        self._notify = agent_base.AgentBaseNotification(sc)
        super(VpnaasRpcSender, self).__init__(None, None)

    def get_vpn_services(self, context, ids=None, filters=None):
        """Gets list of vpnservices for tenant.
        :param context: dictionary which holds details of vpn service type like
            For IPSEC connections :
                List of vpnservices
                lIst of ipsec connections
                ike policy & ipsec policy.
        :param ids: based on which the filter library extracts the data.
        :param filter: based on which the filter library extracts the data.

        Returns: Dictionary of vpn service type which matches with the filters.
        """
        LOG.info(_LI("Sending RPC for GET VPN SERVICES with %(filters)s "),
                 {'filters': filters})
        return self.call(
            context,
            self.make_msg('get_vpn_services', ids=ids, filters=filters))

    def get_vpn_servicecontext(self, context, filters=None):
        """Get list of vpnservice context on this host.
        :param context: dictionary which holds details of vpn service type like
            For IPSEC connections :
                List of vpnservices
                lIst of ipsec connections
                ike policy & ipsec policy.
        :param filter: based on which the filter library extracts the data
        from context dictionary.

        Returns: dictionary of vpnservice
        """
        LOG.info(_LI("Sending RPC for GET VPN SERVICECONTEXT with "
                     "Filters:%(filters)s "),
                 {'filters': filters})
        return self.call(
            context,
            self.make_msg(
                'get_vpn_servicecontext', filters=filters))

    def get_ipsec_conns(self, context, filters):
        """
        Get list of ipsec conns with filters
        specified.
        """
        LOG.info(_LI("Sending RPC for GET IPSEC CONNS with Filters:"
                     "%(filters)s "),
                 {'filters': filters})
        return self.call(
            context,
            self.make_msg(
                'get_ipsec_conns',
                filters=filters))

    def update_status(self, context, status):
        """Update local status.

        This method call updates status attribute of
        VPNServices.
        """
        msg = {'info': {'service_type': const.SERVICE_TYPE,
                        'context': context['agent_info']['context']},
               'notification': [{
                   'resource': context['agent_info']['resource'],
                   'data': {'status': status,
                            'notification_type': (
                                'update_status')}}]
               }
        LOG.info(_LI("Sending Notification 'Update Status' with "
                     "status:%(status)s "),
                 {'status': status})
        self._notify._notification(msg)

    def ipsec_site_conn_deleted(self, context, resource_id):
        """ Notify VPNaaS plugin about delete of ipsec-site-conn """

        msg = {'info': {'service_type': const.SERVICE_TYPE,
                        'context': context['agent_info']['context']},
               'notification': [{
                   'resource': context['agent_info']['resource'],
                   'data': {'resource_id': resource_id,
                            'notification_type': (
                                'ipsec_site_conn_deleted')}}]
               }
        LOG.info(_LI("Sending Notification 'Ipsec Site Conn Deleted' "
                     "for resource:%(resource_id)s "),
                 {'resource_id': resource_id})
        self._notify._notification(msg)


class VPNaasRpcManager(agent_base.AgentBaseRPCManager):
    """
    Implements VPNaasRpcManager class which receives requests
        from Configurator to Agent.

    Methods of this class are invoked by the configurator. Events are
    created according to the requests received and enqueued to worker queues.

    """

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, conf, sc):
        """Instantiates child and parent class objects.

        Passes the instances of core service controller and oslo configuration
        to parent instance in order to provide event enqueue facility for batch
        processing event.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        super(VPNaasRpcManager, self).__init__(sc, conf)

    def vpnservice_updated(self, context, **resource_data):
        """Registers the VPNaas plugin events to update the vpn configurations.

        :param context: dictionary, confined to the specific service type.
        :param resource_data: dictionary, confined to the specific
               operation type.

        Returns: None
        """

        LOG.info(_LI("Received request 'VPN Service Updated'."
                     "for API '%(api)s'"),
                 {'api': resource_data.get('reason', '')})
        arg_dict = {'context': context,
                    'resource_data': resource_data}
        # Serializing the event because simultaneous configure
        # requests overrides the same crypto-map in the service VM
        # which results in corrupting the crypto-map

        resource_type = resource_data.get('rsrc_type')
        if resource_type and resource_type.lower() == 'ipsec_site_connection':
            ev = self.sc.new_event(id='VPNSERVICE_UPDATED',
                                   key=resource_data['resource']['id'],
                                   data=arg_dict,
                                   serialize=True,
                                   binding_key=resource_data[
                                       'resource']['vpnservice_id'])
            msg = "serializing event: %s" % ('VPNSERVICE_UPDATED')
            LOG.debug(msg)
        else:
            ev = self.sc.new_event(id='VPNSERVICE_UPDATED', data=arg_dict)

        self.sc.post_event(ev)


class VPNaasEventHandler(nfp_api.NfpEventHandler):
    """
    Handler class to invoke the vpn driver methods.
    For every event that gets invoked from worker process lands over here
    to make a call to the driver methods.
    """

    def __init__(self, sc, drivers):
        """ Instantiates class object.

        :param sc: Service Controller object that is used to communicate
        with process model core file.
        :param drivers: dictionary of driver name to object mapping

        """
        self._sc = sc
        self._drivers = drivers
        self._plugin_rpc = VpnaasRpcSender(self._sc)

    def _get_driver(self, service_vendor, service_feature):

        driver_id = const.SERVICE_TYPE + service_vendor + service_feature
        return self._drivers[driver_id]

    def handle_event(self, ev):
        """
        Demultiplexes the vpn request to appropriate  driver methods.

        :param ev: event object sent from the process model.

        Returns: None
        """
        if ev.id == 'VPN_SYNC':
            self._sc.poll_event(ev)

        if ev.id == 'VPNSERVICE_UPDATED':
            try:
                msg = ("Worker process with ID: %s starting "
                       "to handle task: %s of topic: %s. "
                       % (os.getpid(),
                          ev.id, const.VPN_GENERIC_CONFIG_RPC_TOPIC))
                LOG.debug(msg)
                agent_info = ev.data['context']['agent_info']
                service_vendor = agent_info['service_vendor']
                service_feature = agent_info['service_feature']
                driver = self._get_driver(service_vendor, service_feature)
                LOG.info(_LI("Invoking driver with service vendor:"
                             "%(service_vendor)s "),
                         {'service_vendor': service_vendor})
                setattr(VPNaasEventHandler, "service_driver", driver)
                self._vpnservice_updated(ev, driver)
            except Exception as err:
                msg = ("Failed to perform the operation: %s. %s"
                       % (ev.id, str(err).capitalize()))
                LOG.error(msg)
            finally:
                self._sc.event_complete(ev)

    def _vpnservice_updated(self, ev, driver):
        """
        Makes call to the respective operation method of vpn driver.

        :param ev: event object sent from the process model.
        :param driver: vpn driver class object.

        Returns: None.
        """
        context = ev.data.get('context')
        resource_data = ev.data.get('resource_data')
        msg = "Vpn service updated from server side"
        LOG.info(msg)

        try:
            driver.vpnservice_updated(context, resource_data)

            if 'ipsec_site_conns' in context['service_info']:
                for item in context['service_info']['ipsec_site_conns']:
                    if item['id'] == resource_data['resource']['id'] and (
                            resource_data['reason'] == 'create'):
                        item['status'] = 'INIT'
                        arg_dict = {'context': context,
                                    'resource_data': resource_data}
                        ev1 = self._sc.new_event(id='VPN_SYNC',
                                                 key='VPN_SYNC', data=arg_dict)
                        self._sc.post_event(ev1)
                    break
        except Exception as err:
            msg = ("Failed to update VPN service. %s"
                   % str(err).capitalize())
            LOG.error(msg)
        reason = resource_data.get('reason')
        rsrc = resource_data.get('rsrc_type')

        if (reason == 'delete' and rsrc == 'ipsec_site_connection'):
            conn = resource_data['resource']
            resource_id = conn['id']
            self._plugin_rpc.ipsec_site_conn_deleted(context,
                                                     resource_id=resource_id)

    def _get_service_vendor(self, vpn_svc):
        """
        Extracts the vendor from the description.
        :param vpn_svc: vpn service operation type dictionary,
        which it gets from filter library

        Returns: None
        """
        svc_desc = vpn_svc['description']
        tokens = svc_desc.split(';')
        vendor = tokens[5].split('=')[1]
        return vendor

    def _sync_ipsec_conns(self, context, svc_context):
        """
        Gets the status of the vpn service.
        :param context: Dictionary of the vpn service type.
        :param vendor: vendor name
        :param svc_context: vpn service operation type dictionary,
        which it gets filter library

        Returns: None
        """
        try:
            return self.service_driver.check_status(context, svc_context)
        except Exception as err:
            msg = ("Failed to sync ipsec connection information. %s."
                   % str(err).capitalize())
            LOG.error(msg)

    @nfp_api.poll_event_desc(event='VPN_SYNC', spacing=10)
    def sync(self, ev):
        """Periodically updates the status of vpn service, whether the
        tunnel is UP or DOWN.

        :param context: Dictionary of the vpn service type.

        Returns: None
        """

        context = ev.data.get('context')
        s2s_contexts = self._plugin_rpc.get_vpn_servicecontext(context)
        state = self._sync_ipsec_conns(context, s2s_contexts[0])
        if state in {const.STATE_ACTIVE,
                     const.STATE_ERROR}:
            return {'poll': False}


def events_init(sc, drivers):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param sc: Object of Service Controller from the process model to regiters
    the different events
    :param drivers: Driver instance registered with the service agent

    Returns: None
    """
    evs = [
        nfp_event.Event(id='VPNSERVICE_UPDATED',
                        handler=VPNaasEventHandler(sc, drivers)),
        nfp_event.Event(id='VPN_SYNC',
                        handler=VPNaasEventHandler(sc, drivers))]

    sc.register_events(evs)


def load_drivers(sc, conf):
    """Loads the drivers dynamically.

    Loads the drivers that register with the agents.
    :param sc: Object of the Service Controller class from core
    service controller.

    Returns: dictionary of instances of the respective driver classes.
    """

    ld = utils.ConfiguratorUtils(conf)
    drivers = ld.load_drivers(const.SERVICE_TYPE)

    for service_type, driver_name in six.iteritems(drivers):
        driver_obj = driver_name(conf=conf)
        drivers[service_type] = driver_obj

    return drivers


def register_service_agent(cm, sc, conf):
    """Registers the agents with Cofigurator module.
    Puts all the agents into the dictionary with their service types.
    :prarm cm: Configurator module's object to communicate back and forth
    :param sc: Object of the Service Controller class from core
    service controller.
    :param conf: Object of oslo configurator passed from the core service
    controller

    Returns: None
    """

    rpcmgr = VPNaasRpcManager(conf, sc)
    cm.register_service_agent(const.SERVICE_TYPE, rpcmgr)


def init_agent(cm, sc, conf):
    """Loads the drivers and registers the agents.
    Loads the dynamicaaly both the drivers and agents, registers the agents
    with their service types.

    :prarm cm: Configurator module's object to communicate back and forth
    :param sc: Object of the Service Controller class from core
    service controller.
    :param conf: Object of oslo configurator passed from the core service
    controller

    Returns: None

    """
    try:
        drivers = load_drivers(sc, conf)
    except Exception as err:
        msg = ("VPNaas failed to load drivers. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = "VPNaas loaded drivers successfully."
        LOG.debug(msg)

    try:
        events_init(sc, drivers)
    except Exception as err:
        msg = ("VPNaas Events initialization unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = "VPNaas Events initialization successful."
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf)
        bdobj = base_driver.BaseDriver(conf)
        bdobj.register_agent_object_with_driver('agent', VpnaasRpcSender(sc))
    except Exception as err:
        msg = ("VPNaas service agent registration unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = "VPNaas service agent registration successful."
        LOG.debug(msg)

    msg = "VPN as a Service Module Initialized."
    LOG.info(msg)


def init_agent_complete(cm, sc, conf):
    """
    Initializes periodic tasks.
    """
    msg = " vpn agent init complete"
    LOG.info(msg)
